"""
App enums.
"""

# pylint: disable=too-many-public-methods, unused-import

import typing as t
import typing_extensions as te  # noqa: F401
import typing_extensions as te
from composio.client.enums._action import Action
from composio.client.enums.base import APPS_CACHE, AppData, _AnnotatedEnum, enum

@enum
class App(_AnnotatedEnum[AppData], path=APPS_CACHE):
    """Class to represent `App` entity."""
    ABLY: 'App'
    ACCELO: 'App'
    ACTIVE_CAMPAIGN: 'App'
    ADOBE: 'App'
    AERO_WORKFLOW: 'App'
    AFFINITY: 'App'
    AGENCYZOOM: 'App'
    AIRTABLE: 'App'
    ALCHEMY: 'App'
    ALTOVIZ: 'App'
    AMAZON: 'App'
    AMCARDS: 'App'
    AMPLITUDE: 'App'
    ANTHROPIC: 'App'
    APPDRAG: 'App'
    APPSFLYER: 'App'
    ASANA: 'App'
    ASHBY: 'App'
    ATLASSIAN: 'App'
    ATTIO: 'App'
    AUTH0: 'App'
    AXONAUT: 'App'
    BAMBOOHR: 'App'
    BANNERBEAR: 'App'
    BASEROW: 'App'
    BATTLENET: 'App'
    BEEMINDER: 'App'
    BITBUCKET: 'App'
    BITWARDEN: 'App'
    BLACKBAUD: 'App'
    BLACKBOARD: 'App'
    BOLDSIGN: 'App'
    BOTBABA: 'App'
    BOX: 'App'
    BRAINTREE: 'App'
    BRANDFETCH: 'App'
    BREEZY_HR: 'App'
    BREVO: 'App'
    BREX: 'App'
    BREX_STAGING: 'App'
    BRIGHTPEARL: 'App'
    BROWSEAI: 'App'
    BROWSERBASE_TOOL: 'App'
    BROWSERHUB: 'App'
    BROWSER_TOOL: 'App'
    BUBBLE: 'App'
    CAL: 'App'
    CALENDLY: 'App'
    CANVA: 'App'
    CANVAS: 'App'
    CHATWORK: 'App'
    CHMEETINGS: 'App'
    CLICKUP: 'App'
    CLOSE: 'App'
    CLOUDFLARE: 'App'
    CODEINTERPRETER: 'App'
    CODE_ANALYSIS_TOOL: 'App'
    CODE_FORMAT_TOOL: 'App'
    COINMARKETCAL: 'App'
    COMPOSIO: 'App'
    CONTENTFUL: 'App'
    CUSTOMER_IO: 'App'
    DAILYBOT: 'App'
    DATADOG: 'App'
    DATAGMA: 'App'
    DATAROBOT: 'App'
    DEEL: 'App'
    DEMIO: 'App'
    DIGICERT: 'App'
    DISCORD: 'App'
    DISCORDBOT: 'App'
    DOCMOSIS: 'App'
    DOCUSIGN: 'App'
    DROPBOX: 'App'
    DROPBOX_SIGN: 'App'
    DYNAMICS365: 'App'
    ECHTPOST: 'App'
    ELEVENLABS: 'App'
    EMBED_TOOL: 'App'
    ENTELLIGENCE: 'App'
    EPIC_GAMES: 'App'
    EVENTBRITE: 'App'
    EXA: 'App'
    EXIST: 'App'
    FACEBOOK: 'App'
    FACTORIAL: 'App'
    FIGMA: 'App'
    FILETOOL: 'App'
    FINAGE: 'App'
    FIRECRAWL: 'App'
    FITBIT: 'App'
    FLUTTERWAVE: 'App'
    FOMO: 'App'
    FORMCARRY: 'App'
    FORMSITE: 'App'
    FRESHBOOKS: 'App'
    FRESHDESK: 'App'
    FRONT: 'App'
    GIT: 'App'
    GITHUB: 'App'
    GMAIL: 'App'
    GOOGLEBIGQUERY: 'App'
    GOOGLECALENDAR: 'App'
    GOOGLEDOCS: 'App'
    GOOGLEDRIVE: 'App'
    GOOGLEMEET: 'App'
    GOOGLESHEETS: 'App'
    GOOGLETASKS: 'App'
    GOOGLE_ANALYTICS: 'App'
    GORGIAS: 'App'
    GO_TO_WEBINAR: 'App'
    GREPTILE: 'App'
    GUMROAD: 'App'
    GURU: 'App'
    HACKERNEWS: 'App'
    HACKERRANK_WORK: 'App'
    HARVEST: 'App'
    HELCIM: 'App'
    HEYGEN: 'App'
    HIGHLEVEL: 'App'
    HISTORY_FETCHER: 'App'
    HUBSPOT: 'App'
    HUMANLOOP: 'App'
    ICIMS_TALENT_CLOUD: 'App'
    IDEA_SCALE: 'App'
    IMAGE_ANALYSER: 'App'
    INDUCED_AI: 'App'
    INTERCOM: 'App'
    INTERZOID: 'App'
    JIRA: 'App'
    JUNGLESCOUT: 'App'
    KEAP: 'App'
    KLAVIYO: 'App'
    KLIPFOLIO: 'App'
    LASTPASS: 'App'
    LAUNCH_DARKLY: 'App'
    LEVER: 'App'
    LEVER_SANDBOX: 'App'
    LEXOFFICE: 'App'
    LINEAR: 'App'
    LINKEDIN: 'App'
    LINKHUT: 'App'
    LISTENNOTES: 'App'
    LMNT: 'App'
    MAILCHIMP: 'App'
    MAILERLITE: 'App'
    MAINTAINX: 'App'
    MATHEMATICAL: 'App'
    MBOUM: 'App'
    MEM0: 'App'
    METATEXTAI: 'App'
    MICROSOFT_CLARITY: 'App'
    MICROSOFT_TEAMS: 'App'
    MICROSOFT_TENANT: 'App'
    MIRO: 'App'
    MIXPANEL: 'App'
    MOCEAN: 'App'
    MONDAY: 'App'
    MOPINION: 'App'
    MORE_TREES: 'App'
    MOXIE: 'App'
    MOZ: 'App'
    MULTIONAI: 'App'
    MURAL: 'App'
    NCSCALE: 'App'
    NETSUITE: 'App'
    NGROK: 'App'
    NOTION: 'App'
    ONCEHUB: 'App'
    ONE_DRIVE: 'App'
    OUTLOOK: 'App'
    PANDADOC: 'App'
    PERPLEXITYAI: 'App'
    PIGGY: 'App'
    PIPEDRIVE: 'App'
    PLACEKEY: 'App'
    POSTHOG: 'App'
    PRECORO: 'App'
    PRINTNODE: 'App'
    PROCESS_STREET: 'App'
    PRODUCTBOARD: 'App'
    QUALAROO: 'App'
    RAFFLYS: 'App'
    RAGTOOL: 'App'
    RAVENSEOTOOLS: 'App'
    REDDIT: 'App'
    RING_CENTRAL: 'App'
    RIPPLING: 'App'
    ROCKET_REACH: 'App'
    SAGE: 'App'
    SALESFORCE: 'App'
    SCREENSHOTONE: 'App'
    SEISMIC: 'App'
    SENDGRID: 'App'
    SERPAPI: 'App'
    SERVICEM8: 'App'
    SHARE_POINT: 'App'
    SHELLTOOL: 'App'
    SHOPIFY: 'App'
    SHORTCUT: 'App'
    SIMPLESAT: 'App'
    SLACK: 'App'
    SLACKBOT: 'App'
    SMARTRECRUITERS: 'App'
    SMUGMUG: 'App'
    SNOWFLAKE: 'App'
    SPIDERTOOL: 'App'
    SQLTOOL: 'App'
    SQUARE: 'App'
    STACK_EXCHANGE: 'App'
    SUPABASE: 'App'
    SURVEY_MONKEY: 'App'
    TAPFORM: 'App'
    TAVILY: 'App'
    TERMINUS: 'App'
    TEXTRAZOR: 'App'
    TIMECAMP: 'App'
    TIMEKIT: 'App'
    TIMELY: 'App'
    TINYPNG: 'App'
    TINYURL: 'App'
    TISANE: 'App'
    TODOIST: 'App'
    TONEDEN: 'App'
    TRELLO: 'App'
    TWITCH: 'App'
    TWITTER: 'App'
    TWITTER_MEDIA: 'App'
    VENLY: 'App'
    VERO: 'App'
    VISME: 'App'
    WABOXAPP: 'App'
    WAKATIME: 'App'
    WAVE_ACCOUNTING: 'App'
    WEATHERMAP: 'App'
    WEBEX: 'App'
    WEBFLOW: 'App'
    WEBTOOL: 'App'
    WIZ: 'App'
    WORKIOM: 'App'
    WORKSPACE_TOOL: 'App'
    XERO: 'App'
    YANDEX: 'App'
    YNAB: 'App'
    YOUSEARCH: 'App'
    YOUTUBE: 'App'
    ZENDESK: 'App'
    ZENSERP: 'App'
    ZEPTOOL: 'App'
    ZOHO: 'App'
    ZOHO_BIGIN: 'App'
    ZOHO_BOOKS: 'App'
    ZOHO_DESK: 'App'
    ZOHO_INVENTORY: 'App'
    ZOHO_INVOICE: 'App'
    ZOHO_MAIL: 'App'
    ZOOM: 'App'

    @property
    def is_local(self) -> bool:
        """The tool is local if set to `True`"""
        return self.load().is_local

    @property
    def name(self) -> str:
        """Name of the app."""
        return self.load().name

    def get_actions(self, tags: t.Optional[t.List[str]]=None) -> t.Iterator[Action]:
        """
        Get actions for the given app filtered by the `tags`

        :param tags: List of tags to filter the actions
        :return: Iterator object which yields `Action`
        """
        tags = tags or []
        app = f'{self.slug.lower()}_'
        for action in Action.all():
            if not action.slug.lower().startswith(app):
                continue
            if len(tags) == 0:
                yield action
            if any((tag in action.tags for tag in tags)):
                yield action


import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class DeleteDependencyRequest(BaseModel):
    """Request schema for `DeleteDependency`"""

    task_id: str = Field(
        ...,
        alias="task_id",
        description="",
    )
    depends_on: str = Field(
        ...,
        alias="depends_on",
        description="",
    )
    dependency_of: str = Field(
        ...,
        alias="dependency_of",
        description="",
    )
    custom_task_ids: bool = Field(
        default=None,
        alias="custom_task_ids",
        description=(
            'If you want to reference a task by it"s custom task id, this value must '
            "be `true`. "
        ),
    )
    team_id: int = Field(
        default=None,
        alias="team_id",
        description=(
            "Only used when the `custom_task_ids` parameter is set to `true`.   For example: "
            "`custom_task_ids=true&team_id=123`. "
        ),
    )


class DeleteDependencyResponse(BaseModel):
    """Response schema for `DeleteDependency`"""

    data: t.Dict[str, t.Any]


class DeleteDependency(OpenAPIAction):
    """Remove the dependency relationship between two or more tasks."""

    _tags = ["Task Relationships"]
    _display_name = "delete_dependency"
    _request_schema = DeleteDependencyRequest
    _response_schema = DeleteDependencyResponse

    url = "https://api.clickup.com/api/v2"
    path = "/task/{task_id}/dependency"
    method = "delete"
    operation_id = "TaskRelationships_removeDependency"
    action_identifier = "/task/{task_id}/dependency_delete"

    path_params = {"task_id": "task_id"}
    query_params = {
        "depends_on": "depends_on",
        "dependency_of": "dependency_of",
        "custom_task_ids": "custom_task_ids",
        "team_id": "team_id",
    }
    header_params = {}
    request_params = {}

    aliases = {}



import ssl
from typing import Dict
from urllib.request import Request, urlopen

from pydantic import BaseModel, Field

from composio.tools.base.local import LocalAction


# pylint: disable=import-outside-toplevel
class ScrapeWebsiteElementToolRequest(BaseModel):
    website_url: str = Field(..., description="Mandatory website url to read the file")
    element_selector: str = Field(
        ..., description="CSS selector for the element to scrape"
    )


class ScrapeWebsiteElementToolResponse(BaseModel):
    element_content: str = Field(..., description="The content of the selected element")


class ScrapeWebsiteElement(
    LocalAction[
        ScrapeWebsiteElementToolRequest,
        ScrapeWebsiteElementToolResponse,
    ]
):
    """Scrape website element"""

    _tags = ["Web browser"]

    def execute(
        self,
        request: ScrapeWebsiteElementToolRequest,
        metadata: Dict,
    ) -> ScrapeWebsiteElementToolResponse:
        """Scrape a specific element from the website and return its content"""
        try:
            from bs4 import BeautifulSoup
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError("Failed to import BeautifulSoup:", e) from e

        req = Request(
            request.website_url,
            headers={
                "User-Agent": (
                    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
                    "AppleWebKit/537.36 (KHTML, like Gecko) "
                    "Chrome/58.0.3029.110 Safari/537.3"
                )
            },
        )
        context = ssl.create_default_context()
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
        with urlopen(req, context=context) as response:
            html = response.read().decode("utf-8")
            soup = BeautifulSoup(html, "html.parser")
            element = soup.select_one(request.element_selector)
            if element:
                return ScrapeWebsiteElementToolResponse(element_content=str(element))
        raise ValueError(
            f"No content found for element selector {request.element_selector}"
        )



# Import necessary libraries
import math
import os
import shutil
from collections import Counter, defaultdict, namedtuple
from importlib import resources
from pathlib import Path

from composio.tools.local.base.utils.grep_ast import TreeContext
from composio.tools.local.base.utils.parser import filename_to_lang
from composio.tools.local.base.utils.utils import (
    get_mtime,
    get_rel_fname,
    print_if_verbose,
    token_count,
)


# Define a named tuple for storing tag information
Tag = namedtuple("Tag", ["rel_fname", "fname", "line", "name", "kind"])

"""
RepoMap: Generates a structured view of a code repository.

Key components:
1. chat_fnames: Files active in the current conversation.
   - Given high personalization priority in ranking.
   - Excluded from final repo map output.

2. mentioned_fnames: Files of particular interest.
   - Given medium personalization priority in ranking.
   - Included in final repo map output.

3. other_fnames: All other repository files.
   - Given low personalization priority.
   - Included in output if relevant or space allows.

Process:
1. Collect tags (definitions and references) from all files.
2. Build a graph representing relationships between files and identifiers.
3. Use PageRank with personalization to rank files and identifiers.
4. Generate a tree-like structure of the most important elements.
5. Optimize the output to fit within a specified token limit.

The resulting repo map provides a context-aware overview of the repository,
emphasizing relevant files and code structures based on the current conversation
and mentioned points of interest.
"""


class RepoMap:
    # Class variables for caching
    CACHE_VERSION = 1
    TAGS_CACHE_DIR = f".composio.tags.cache.v{CACHE_VERSION}"

    cache_missing = False
    warned_files: set[str] = set()

    def __init__(
        self,
        map_tokens=10240,
        root=None,
        repo_content_prefix=None,
        verbose=False,
        max_context_window=10000,
    ):
        """
        Initialize the RepoMap object.

        :param map_tokens: Maximum number of tokens for the repo map
        :param root: Root directory of the repository
        :param repo_content_prefix: Prefix for repo content
        :param verbose: Enable verbose output
        :param max_context_window: Maximum context window size
        """

        self.verbose = verbose

        # Set root directory
        if not root:
            root = os.getcwd()
        self.root = root

        # Load tags cache
        self.load_tags_cache()

        self.max_map_tokens = map_tokens
        self.max_context_window = max_context_window
        self.TAGS_CACHE = None
        self.tree_cache = {}
        self.token_count = token_count
        self.repo_content_prefix = repo_content_prefix

    def get_repo_map(
        self, chat_files, other_files, mentioned_fnames=None, mentioned_idents=None
    ):
        """
        Generate a repository map based on the given files and mentions.

        :param chat_files: Files currently in the chat
        :param other_files: Other files in the repository
        :param mentioned_fnames: Mentioned file names
        :param mentioned_idents: Mentioned identifiers
        :return: Repository map as a string
        """
        # Early exit conditions
        if self.max_map_tokens <= 0 or not other_files:
            print_if_verbose(
                "Exiting repo-map due to max_map_tokens <= 0 or no other_files",
                self.verbose,
            )
            return "error"

        # Initialize mentioned sets if not provided
        mentioned_fnames = mentioned_fnames or set()
        mentioned_idents = mentioned_idents or set()

        max_map_tokens = self.max_map_tokens

        # Adjust max_map_tokens when no files are in the chat
        MUL = 8
        padding = 4096
        if max_map_tokens and self.max_context_window:
            target = min(max_map_tokens * MUL, self.max_context_window - padding)
        else:
            target = 0
        if not chat_files and self.max_context_window and target > 0:
            max_map_tokens = target

        try:
            # Generate ranked tags map
            files_listing = self.get_ranked_tags_map(
                chat_files,
                other_files,
                max_map_tokens,
                mentioned_fnames,
                mentioned_idents,
            )
        except RecursionError:
            # Handle recursion error (possibly due to large git repo)
            self.max_map_tokens = 0
            print_if_verbose("Exiting repo-map due to RecursionError", self.verbose)
            return "error"

        if not files_listing:
            print_if_verbose(
                "Exiting repo-map due to empty files_listing", self.verbose
            )
            return "error"

        # Count tokens in the files listing
        num_tokens = self.token_count(files_listing)
        print_if_verbose(f"Repo-map: {num_tokens / 1024:.1f} k-tokens", self.verbose)

        # Prepare repo content string
        other = "other " if chat_files else ""
        if self.repo_content_prefix:
            repo_content = self.repo_content_prefix.format(other=other)
        else:
            repo_content = ""

        repo_content += files_listing

        return repo_content

    def load_tags_cache(self):
        from diskcache import Cache  # pylint: disable=C0415

        path = Path(self.root) / self.TAGS_CACHE_DIR
        if not path.exists():
            self.cache_missing = True
        self.TAGS_CACHE = Cache(path)

    def get_tags(self, fname: str, rel_fname: str) -> list[Tag]:
        """
        Get tags for a file, using cache if available.

        :param fname: Absolute file name
        :param rel_fname: Relative file name
        :return: List of tags
        """
        file_mtime = get_mtime(fname)
        if file_mtime is None:
            print_if_verbose(
                f"Warning: Unable to get modification time for {fname}, skipping",
                self.verbose,
            )
            return []

        if self.TAGS_CACHE is None:
            print_if_verbose(
                "Warning: Tags cache is not initialized, something went wrong",
                self.verbose,
            )
            return []

        cache_key = fname
        cache_data = self.TAGS_CACHE.get(cache_key)

        if cache_data is not None and cache_data.get("mtime") == file_mtime:
            return cache_data.get("data", [])

        # Cache miss or outdated: generate new tags
        data = list(self.get_tags_raw(fname, rel_fname))

        # Update cache
        self.TAGS_CACHE.set(cache_key, {"mtime": file_mtime, "data": data})
        return data

    def get_tags_raw(self, fname, rel_fname):
        """
        Generate tags for a file using tree-sitter and pygments.

        :param fname: Absolute file name
        :param rel_fname: Relative file name
        :yield: Tag objects
        """
        lang = filename_to_lang(fname)
        if not lang:
            print_if_verbose(
                "Exiting get_tags_raw due to no language detected", self.verbose
            )
            return

        from tree_sitter_languages import (  # pylint: disable=C0415
            get_language,
            get_parser,
        )

        language = get_language(lang)
        parser = get_parser(lang)

        # Load tags query
        try:
            scm_fname = resources.files(__package__).joinpath(
                "queries", f"tree-sitter-{lang}-tags.scm"
            )
        except KeyError:
            print_if_verbose(
                "Exiting get_tags_raw due to KeyError in loading tags query",
                self.verbose,
            )
            return

        if not scm_fname.is_file():
            print_if_verbose(
                "Exiting get_tags_raw due to non-existent query_scm", self.verbose
            )
            return
        query_scm = scm_fname.read_text()

        # Parse code
        with open(fname, "r", encoding="utf-8") as file:
            code = file.read().strip()

        if not code:
            print_if_verbose("Exiting get_tags_raw due to empty code", self.verbose)
            return

        tree = parser.parse(bytes(code, "utf-8"))

        # Run tags query
        query = language.query(query_scm)
        captures = query.captures(tree.root_node)

        saw = set()
        for node, tag in captures:
            if tag.startswith("name.definition."):
                kind = "def"
            elif tag.startswith("name.reference."):
                kind = "ref"
            else:
                continue
            saw.add(kind)

            yield Tag(
                rel_fname=rel_fname,
                fname=fname,
                name=node.text.decode("utf-8"),
                kind=kind,
                line=node.start_point[0],
            )

        # If no references found, use pygments for additional tagging
        if "ref" in saw:
            print_if_verbose(
                "Exiting get_tags_raw after processing references", self.verbose
            )
            print_if_verbose(fname, self.verbose)
            return
        if "def" not in saw:
            print_if_verbose(
                "Exiting get_tags_raw due to no definitions found", self.verbose
            )
            return

        from pygments.util import ClassNotFound  # pylint: disable=C0415

        try:
            from pygments.lexers import (  # pylint: disable=C0415
                guess_lexer_for_filename,
            )
            from pygments.token import Token  # pylint: disable=C0415

            lexer = guess_lexer_for_filename(fname, code)
            tokens = [
                token[1] for token in lexer.get_tokens(code) if token[0] in Token.Name
            ]
        except ClassNotFound:
            print_if_verbose(
                "Exiting get_tags_raw due to ClassNotFound in lexer guessing",
                self.verbose,
            )
            tokens = code.split()

        for token in tokens:
            yield Tag(
                rel_fname=rel_fname,
                fname=fname,
                name=token,
                kind="ref",
                line=-1,
            )

    def get_ranked_tags(
        self, chat_fnames, other_fnames, mentioned_fnames, mentioned_idents
    ):  # pylint: disable=R0915
        """
        Generate ranked tags for files in the repository.

        :param chat_fnames: Files currently in the chat
        :param other_fnames: Other files in the repository
        :param mentioned_fnames: Mentioned file names
        :param mentioned_idents: Mentioned identifiers
        :return: List of ranked tags
        """
        import networkx as nx  # pylint: disable=C0415

        defines = defaultdict(set)
        references = defaultdict(list)
        definitions = defaultdict(set)

        personalization = {}
        fnames = sorted(set(chat_fnames).union(set(other_fnames)))
        chat_rel_fnames = set()

        # Improved personalization logic
        chat_weight = 10.0
        mentioned_weight = 5.0
        other_weight = 1.0

        total_weight = (
            len(chat_fnames) * chat_weight
            + len(mentioned_fnames) * mentioned_weight
            + len(other_fnames) * other_weight
        )
        self.cache_missing = False

        # Process each file
        for fname in fnames:
            if not Path(fname).is_file():
                if fname not in self.warned_files:
                    if Path(fname).exists():
                        print_if_verbose(
                            f"Repo-map can't include {fname}, it is not a normal file",
                            self.verbose,
                        )
                    else:
                        print_if_verbose(
                            f"Repo-map can't include {fname}, it no longer exists",
                            self.verbose,
                        )
                self.warned_files.add(fname)
                continue

            rel_fname = get_rel_fname(self.root, fname)
            if fname in chat_fnames:
                personalization[rel_fname] = chat_weight / total_weight
                chat_rel_fnames.add(rel_fname)
            elif rel_fname in mentioned_fnames:
                personalization[rel_fname] = mentioned_weight / total_weight
            else:
                personalization[rel_fname] = other_weight / total_weight

            tags = self.get_tags(fname, rel_fname)
            if tags is None:
                continue

            for tag in tags:
                if tag.kind == "def":
                    defines[tag.name].add(rel_fname)
                    key = (rel_fname, tag.name)
                    definitions[key].add(tag)
                if tag.kind == "ref":
                    references[tag.name].append(rel_fname)

        # If no references, use definitions as references
        if not references:
            references = dict((k, list(v)) for k, v in defines.items())

        idents = set(defines.keys()).intersection(set(references.keys()))

        # Create graph
        G = nx.MultiDiGraph()

        for ident in idents:
            definers = defines[ident]
            mul = (
                2.0
                if ident in mentioned_idents
                else 0.5 if ident.startswith("_") else 1.0
            )

            for referencer, num_refs in Counter(references[ident]).items():
                for definer in definers:
                    # Scale down high-frequency mentions
                    num_refs = math.sqrt(num_refs)
                    G.add_edge(referencer, definer, weight=mul * num_refs, ident=ident)

        # Calculate PageRank
        try:
            ranked = nx.pagerank(G, weight="weight", personalization=personalization)
        except ZeroDivisionError:
            print_if_verbose(
                "Exiting get_ranked_tags due to ZeroDivisionError in PageRank calculation",
                self.verbose,
            )
            return []

        # Distribute rank across edges
        ranked_definitions = defaultdict(float)
        for src in G.nodes:
            src_rank = ranked[src]
            total_weight = sum(
                data["weight"] for _src, _dst, data in G.out_edges(src, data=True)
            )
            for _src, dst, data in G.out_edges(src, data=True):
                data["rank"] = src_rank * data["weight"] / total_weight
                ident = data["ident"]
                ranked_definitions[(dst, ident)] += data["rank"]

        ranked_tags = []
        ranked_definitions = sorted(
            ranked_definitions.items(), reverse=True, key=lambda x: x[1]
        )

        # Generate ranked tags
        for (fname, ident), rank in ranked_definitions:
            if fname not in chat_rel_fnames:
                ranked_tags.extend(definitions.get((fname, ident), []))

        rel_other_fnames_without_tags = set(
            get_rel_fname(self.root, fname) for fname in other_fnames
        )

        fnames_already_included = set(rt[0] for rt in ranked_tags)

        # Add remaining files to ranked tags
        top_rank = sorted(
            [(rank, node) for (node, rank) in ranked.items()], reverse=True
        )
        for rank, fname in top_rank:
            if fname in rel_other_fnames_without_tags:
                rel_other_fnames_without_tags.remove(fname)
            if fname not in fnames_already_included:
                ranked_tags.append((fname,))

        for fname in rel_other_fnames_without_tags:
            ranked_tags.append((fname,))

        return ranked_tags

    def get_ranked_tags_map(
        self,
        chat_fnames,
        other_fnames=None,
        max_map_tokens=None,
        mentioned_fnames=None,
        mentioned_idents=None,
    ):
        """
        Generate a ranked tags map for the repository.

        :param chat_fnames: Files currently in the chat
        :param other_fnames: Other files in the repository
        :param max_map_tokens: Maximum number of tokens for the map
        :param mentioned_fnames: Mentioned file names
        :param mentioned_idents: Mentioned identifiers
        :return: Formatted string of the ranked tags map
        """
        # print("Starting get_ranked_tags_map")
        if not other_fnames:
            other_fnames = []
        if not max_map_tokens:
            max_map_tokens = self.max_map_tokens

        mentioned_fnames = mentioned_fnames or set()
        mentioned_idents = mentioned_idents or set()

        ranked_tags = self.get_ranked_tags(
            chat_fnames, other_fnames, mentioned_fnames, mentioned_idents
        )

        num_tags = len(ranked_tags)
        lower_bound = 0
        upper_bound = num_tags
        best_tree = None
        best_tree_tokens = 0

        chat_rel_fnames = [get_rel_fname(self.root, fname) for fname in chat_fnames]
        middle = min(max_map_tokens // 25, num_tags)

        self.tree_cache = {}

        while lower_bound <= upper_bound:
            tree = self.to_tree(ranked_tags[:middle], chat_rel_fnames)
            num_tokens = self.token_count(tree)

            if best_tree_tokens < num_tokens < max_map_tokens:
                best_tree = tree
                best_tree_tokens = num_tokens

            if num_tokens < max_map_tokens:
                lower_bound = middle + 1
            else:
                upper_bound = middle - 1

            middle = (lower_bound + upper_bound) // 2

        return best_tree

    def render_tree(self, abs_fname, rel_fname, lois):
        key = (rel_fname, tuple(sorted(lois)))

        if key in self.tree_cache:
            return self.tree_cache[key]

        # use python to read the file
        with open(abs_fname, "r", encoding="utf-8") as file:
            code = file.read()
        if not code.endswith("\n"):
            code += "\n"
        context = TreeContext(
            rel_fname,
            code,
            color=False,
            line_number=False,
            child_context=False,
            last_line=False,
            margin=0,
            mark_lois=False,
            loi_pad=0,
            # header_max=30,
            show_top_of_file_parent_scope=False,
        )

        context.add_lines_of_interest(lois)
        context.add_context()
        res = context.format()
        self.tree_cache[key] = res
        return res

    def to_tree(self, tags, chat_rel_fnames):
        if not tags:
            return ""

        tags = [tag for tag in tags if tag[0] not in chat_rel_fnames]
        tags = sorted(tags)

        cur_fname = None
        cur_abs_fname = None
        lois = None
        output = ""

        # add a bogus tag at the end so we trip the this_fname != cur_fname...
        dummy_tag = (None,)
        for tag in tags + [dummy_tag]:
            this_rel_fname = tag[0]

            # ... here ... to output the final real entry in the list
            if this_rel_fname != cur_fname:
                if lois is not None:
                    output += "\n"
                    if cur_fname is not None:
                        output += cur_fname + ":\n"
                    lang = filename_to_lang(cur_abs_fname)
                    if lang:
                        output += self.render_tree(cur_abs_fname, cur_fname, lois)
                    if lang is None:
                        # print("Skipping : ", cur_abs_fname)
                        continue
                    lois = None
                elif cur_fname:
                    output += "\n" + cur_fname + "\n"
                if isinstance(tag, Tag):
                    lois = []
                    cur_abs_fname = tag.fname
                cur_fname = this_rel_fname

            if lois is not None:
                lois.append(tag.line)

        # truncate long lines, in case we get minified js or something else crazy
        output = "\n".join([line[:100] for line in output.splitlines()]) + "\n"

        return output

    def delete_cache(self):
        """Delete the tags cache."""
        cache_path = Path(self.root) / self.TAGS_CACHE_DIR
        # print("Deleting cache: ", cache_path)
        if cache_path.exists():
            # Remove all files and subdirectories
            for item in cache_path.glob("*"):
                if item.is_file():
                    item.unlink()
                elif item.is_dir():
                    shutil.rmtree(item)
            # print(f"Cache contents deleted: {cache_path}")
        else:
            # print("No cache found to delete.")
            from diskcache import Cache  # pylint: disable=C0415

            # Reset the cache object
            self.TAGS_CACHE = Cache(cache_path)
            self.cache_missing = True



"""Base abstractions."""

import hashlib
import inspect
import json
import typing as t
from abc import abstractmethod
from pathlib import Path

import inflection
import jsonref
import pydantic
from pydantic import BaseModel, Field

from composio.client.enums import Action as ActionEnum
from composio.exceptions import ComposioSDKError
from composio.utils.logging import WithLogger
from composio.utils.pydantic import parse_pydantic_error


GroupID = t.Literal["runtime", "local", "api"]
ModelType = t.TypeVar("ModelType")
ActionResponse = t.TypeVar("ActionResponse")
ActionRequest = t.TypeVar("ActionRequest")
Loadable = t.TypeVar("Loadable")
ToolRegistry = t.Dict[GroupID, t.Dict[str, "Tool"]]
ActionsRegistry = t.Dict[GroupID, t.Dict[str, "Action"]]
TriggersRegistry = t.Dict[GroupID, t.Dict[str, t.Any]]

tool_registry: ToolRegistry = {"runtime": {}, "local": {}, "api": {}}
action_registry: ActionsRegistry = {"runtime": {}, "local": {}, "api": {}}
trigger_registry: TriggersRegistry = {"runtime": {}, "local": {}, "api": {}}

DEPRECATED_MARKER = "<<DEPRECATED use "


def remove_json_ref(data: t.Dict) -> t.Dict:
    return json.loads(
        jsonref.dumps(
            jsonref.replace_refs(
                obj=data,
                lazy_load=False,
            ),
            indent=2,
        )
    )


def generate_app_id(name: str) -> str:
    # Generate a 32-character hash using MD5
    hash_string = hashlib.md5(name.encode()).hexdigest()
    # Insert hyphens at the specified positions
    return "-".join(
        (
            hash_string[:8],
            hash_string[8:12],
            hash_string[12:16],
            hash_string[16:20],
            hash_string[20:],
        )
    )


class InvalidClassDefinition(ComposioSDKError):
    """Raise when a class is not defined properly."""


class ExecuteResponse(BaseModel):
    """Execute action response."""


class _Attributes:
    name: str
    """Name representation."""

    enum: str
    """Enum key."""

    display_name: str
    """Display compatible name."""

    description: str
    """Description string."""

    logo: str
    """URL for the resource logo."""


class _Request(t.Generic[ModelType]):
    """Request util."""

    def __init__(self, model: t.Type[ModelType]) -> None:
        """Initialize request model."""
        self.model = model

    def schema(self) -> t.Dict:
        """Build request schema."""
        request = t.cast(t.Type[BaseModel], self.model).model_json_schema(by_alias=True)
        request = remove_json_ref(request)
        if "$defs" in request:
            del request["$defs"]

        properties = request.get("properties", {})
        for prop in properties.values():
            if prop.get("file_readable", False):
                prop["oneOf"] = [
                    {
                        "type": prop.get("type"),
                        "description": prop.get("description", ""),
                    },
                    {
                        "type": "string",
                        "format": "file-path",
                        "description": f"File path to {prop.get('description', '')}",
                    },
                ]
                del prop["type"]  # Remove original type to avoid conflict in oneOf
                continue

            if "allOf" in prop and len(prop["allOf"]) == 1:
                (schema,) = prop.pop("allOf")
                prop.update(schema)
                if "enum" in schema:
                    prop[
                        "description"
                    ] += f" Note: choose value only from following options - {prop['enum']}"

            if "anyOf" in prop:
                typedef, *_ = [
                    td for td in prop["anyOf"] if td.get("type", "null") != "null"
                ]
                prop["type"] = typedef["type"]

        request["properties"] = properties
        return request

    def parse(self, request: t.Dict) -> ModelType:
        """Parse request."""
        try:
            return self.model(**request)
        except pydantic.ValidationError as e:
            raise ValueError(parse_pydantic_error(e)) from e


class _Response(t.Generic[ModelType]):
    """Response util."""

    def __init__(self, model: t.Type[ModelType]) -> None:
        """Initialize request model."""
        self.model = model
        self.wrapper = self.wrap(model=model)

    @classmethod
    def wrap(cls, model: t.Type[ModelType]) -> t.Type[BaseModel]:
        class wrapper(model):  # type: ignore
            successful: bool = Field(
                ...,
                description="Whether or not the action execution was successful or not",
            )
            error: t.Optional[str] = Field(
                None,
                description="Error if any occurred during the execution of the action",
            )

        return t.cast(t.Type[BaseModel], wrapper)

    def schema(self) -> t.Dict:
        """Build request schema."""
        schema = self.wrapper.model_json_schema(by_alias=True)
        schema = remove_json_ref(schema)
        if "$defs" in schema:
            del schema["$defs"]

        properties = schema.get("properties", {})
        for prop in properties.values():
            if prop.get("file_readable", False):
                prop["oneOf"] = [
                    {
                        "type": prop.get("type"),
                        "description": prop.get("description", ""),
                    },
                    {
                        "type": "string",
                        "format": "file-path",
                        "description": f"File path to {prop.get('description', '')}",
                    },
                ]
                del prop["type"]  # Remove original type to avoid conflict in oneOf
                continue

            if "allOf" in prop and len(prop["allOf"]) == 1:
                (schema,) = prop.pop("allOf")
                prop.update(schema)
                if "enum" in schema:
                    prop[
                        "description"
                    ] += f" Note: choose value only from following options - {prop['enum']}"

        schema["properties"] = properties
        schema["title"] = self.model.__name__
        return remove_json_ref(schema)


class ActionBuilder:
    @staticmethod
    def get_generics(obj: t.Type["Action"]) -> t.Tuple[t.Any, t.Any]:
        for base in getattr(obj, "__orig_bases__"):
            args = t.get_args(base)
            if len(args) == 2:
                return args  # type: ignore
        raise ValueError("No type generics found")

    @classmethod
    def set_generics(cls, name: str, obj: t.Type["Action"]) -> None:
        try:
            request, response = cls.get_generics(obj=obj)
            if request == ActionRequest or response == ActionResponse:
                raise ValueError(f"Invalid type generics, ({request}, {response})")
        except ValueError as e:
            raise InvalidClassDefinition(
                "Invalid action class definition, please define your class "
                "using request and response type generics; "
                f"class {name}(Action[RequestModel, ResponseModel])"
            ) from e

        setattr(obj, "request", _Request(request))
        setattr(obj, "response", _Response(response))

    @staticmethod
    def validate(name: str, obj: t.Type["Action"]) -> None:
        if getattr(getattr(obj, "execute"), "__isabstractmethod__", False):
            raise InvalidClassDefinition(f"Please implement {name}.execute")

    @classmethod
    def set_metadata(cls, obj: t.Type["Action"]) -> None:
        setattr(obj, "file", getattr(obj, "file", Path(inspect.getfile(obj))))
        setattr(obj, "name", getattr(obj, "name", inflection.underscore(obj.__name__)))
        setattr(
            obj,
            "enum",
            getattr(obj, "enum", inflection.underscore(obj.__name__).upper()),
        )
        setattr(
            obj,
            "display_name",
            getattr(
                obj,
                "display_name",
                inflection.humanize(inflection.underscore(obj.__name__)),
            ),
        )
        setattr(
            obj,
            "description",
            cls._get_description(obj),
        )
        description, *_ = obj.description.split(DEPRECATED_MARKER, maxsplit=1)
        if len(description) > 1024:
            raise InvalidClassDefinition(
                f"Description for action `{obj.__name__}` contains more than 1024 characters"
            )

    @staticmethod
    def _get_description(obj) -> str:
        description = t.cast(
            str,
            (
                (obj.__doc__ if obj.__doc__ else obj.display_name)
                .replace("\n    ", " ")
                .strip()
            ),
        )
        description, separator, enum = description.partition(DEPRECATED_MARKER)
        return inflection.titleize(description) + separator + enum


class ActionMeta(type):
    """Action metaclass."""

    def __init__(  # pylint: disable=unused-argument, self-cls-assignment
        cls,
        name: str,
        bases: t.Tuple,
        dict_: t.Dict,
        abs: bool = False,
    ) -> None:
        """Initialize action class."""
        if abs or name == "Action":
            return

        cls = t.cast(t.Type["Action"], cls)
        ActionBuilder.validate(name=name, obj=cls)
        ActionBuilder.set_generics(name=name, obj=cls)
        ActionBuilder.set_metadata(obj=cls)


class Action(
    WithLogger,
    _Attributes,
    t.Generic[ActionRequest, ActionResponse],
    metaclass=ActionMeta,
):
    """Action abstraction."""

    _tags: t.Optional[t.List[str]] = None

    _schema: t.Optional[t.Dict] = None

    tool: str
    """Toolname."""

    request: _Request[ActionRequest]
    """Request helper."""

    response: _Response[ActionResponse]
    """Response helper."""

    file: str
    """Path to the file containing the action"""

    requires: t.Optional[t.List[str]] = None
    """List of dependencies required to run this action."""

    no_auth: bool = False
    """If set `True`, the action does not require a connected account."""

    def __init_subclass__(cls, abs: bool = False) -> None:
        """Initialize subclas."""

    @classmethod
    def tags(cls) -> t.List[str]:
        """Tags for the given action."""
        return cls._tags or []

    @classmethod
    def _generate_schema(cls) -> None:
        """Generate action schema."""
        cls._schema = {
            "name": cls.name,
            "enum": cls.enum,
            "appName": cls.tool,
            "appId": generate_app_id(cls.tool),
            "tags": cls.tags(),
            "response": cls.response.schema(),
            "parameters": cls.request.schema(),
            "displayName": cls.display_name,
            "description": cls.description,
        }

    @classmethod
    def schema(cls) -> t.Dict:
        """Action schema."""
        if cls._schema is None:
            cls._generate_schema()
        return cls._schema  # type: ignore

    @abstractmethod
    def execute(
        self,
        request: ActionRequest,
        metadata: t.Dict,
    ) -> ActionResponse:
        """Execute the action."""


class ToolBuilder:
    @staticmethod
    def validate(obj: t.Type["Tool"], name: str, methods: t.Tuple[str, ...]) -> None:
        for method in methods:
            if getattr(getattr(obj, method), "__isabstractmethod__", False):
                raise InvalidClassDefinition(f"Please implement {name}.{method}")

            if not inspect.ismethod(getattr(obj, method)):
                raise InvalidClassDefinition(
                    f"Please implement {name}.{method} as class method"
                )

    @staticmethod
    def set_metadata(obj: t.Type["Tool"]) -> None:
        setattr(obj, "file", Path(inspect.getfile(obj)))
        setattr(obj, "gid", getattr(obj, "gid", "local"))
        setattr(obj, "name", getattr(obj, "name", inflection.underscore(obj.__name__)))
        setattr(obj, "enum", getattr(obj, "enum", obj.name).upper())
        setattr(
            obj,
            "display_name",
            getattr(
                obj,
                "display_name",
                inflection.humanize(inflection.underscore(obj.__name__)),
            ),
        )
        setattr(obj, "description", (obj.__doc__ or obj.display_name).lstrip().rstrip())
        setattr(obj, "_actions", getattr(obj, "_actions", {}))
        setattr(obj, "_triggers", getattr(obj, "_triggers", {}))

    @staticmethod
    def setup_children(obj: t.Type["Tool"]) -> None:
        if obj.gid not in action_registry:
            action_registry[obj.gid] = {}

        for action in obj.actions():
            action.tool = obj.name
            action.enum = f"{obj.enum}_{action.name.upper()}"
            if obj.requires is not None:
                action.requires = list(set(obj.requires + (action.requires or [])))
            obj._actions[action.enum] = action  # pylint: disable=protected-access
            action_registry[obj.gid][action.enum] = action  # type: ignore

            if hasattr(obj, "logo"):
                setattr(action, "logo", getattr(obj, "logo"))

        if not hasattr(obj, "triggers"):
            return

        if obj.gid not in trigger_registry:
            trigger_registry[obj.gid] = {}

        for trigger in obj.triggers():  # type: ignore
            trigger.tool = obj.name
            trigger.enum = f"{obj.enum}_{trigger.name.upper()}"
            obj._triggers[trigger.enum] = trigger  # type: ignore  # pylint: disable=protected-access
            trigger_registry[obj.gid][trigger.enum] = trigger  # type: ignore

            if hasattr(obj, "logo"):
                setattr(trigger, "logo", getattr(obj, "logo"))


class Tool(WithLogger, _Attributes):
    """Tool abstraction."""

    gid: GroupID
    """Group ID for this tool."""

    file: Path
    """Path to module file."""

    name: str
    """Tool name."""

    requires: t.Optional[t.List[str]] = None
    """List of dependencies required for running this tool."""

    _schema: t.Optional[t.Dict] = None
    """Schema for the app."""

    _actions: t.Dict[str, t.Type[Action]]
    """Actions container"""

    def __init_subclass__(cls, autoload: bool = False) -> None:
        """Initialize a tool class."""

    @classmethod
    def get(cls, enum: ActionEnum) -> t.Type[Action]:
        """Returns the"""
        return cls._actions[enum.slug]

    @classmethod
    @abstractmethod
    def actions(cls) -> t.List[t.Type[t.Any]]:
        """Get collection of actions for the tool."""

    @classmethod
    def _generate_schema(cls) -> None:
        """Generate schema for the app."""
        cls._schema = {
            "name": cls.name,
            "displayName": cls.display_name,
            "metaData": {
                "toolName": cls.name,
                "groupId": cls.gid,
                "displayName": cls.display_name,
                "description": cls.description,
                "toolPath": str(cls.file),
            },
            "integration": {},
            "description": cls.description,
            "actions": [action.schema() for action in cls.actions()],
        }

    @classmethod
    def schema(cls) -> t.Dict:
        """Get tool schema."""
        if cls._schema is None:
            cls._generate_schema()
        return cls._schema  # type: ignore

    def _load(self, loadable: t.Type[Loadable]) -> Loadable:
        """Load action class."""
        instance = loadable()
        return instance

    def execute(
        self,
        action: str,
        params: t.Dict,
        metadata: t.Optional[t.Dict] = None,
    ) -> t.Dict:
        """
        Execute the given action

        :param action: Name of the action.
        :param params: Execution parameters.
        :param metadata: A dictionary containing metadata for action.
        """
        raise NotImplementedError()

    @classmethod
    def register(cls: t.Type["Tool"]) -> None:
        """Register given tool to the registry."""
        # TODO(Viraj): Check if the tool is already registered
        if cls.gid not in tool_registry:
            tool_registry[cls.gid] = {}
        tool_registry[cls.gid][cls.enum] = cls()



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetAccessTokenRequest(BaseModel):
    """Request schema for `GetAccessToken`"""

    client_id: str = Field(
        ...,
        alias="client_id",
        description="Oauth app client id",
    )
    client_secret: str = Field(
        ...,
        alias="client_secret",
        description="Oauth app client secret",
    )
    code: str = Field(
        ...,
        alias="code",
        description="Code given in redirect url",
    )


class GetAccessTokenResponse(BaseModel):
    """Response schema for `GetAccessToken`"""

    data: t.Dict[str, t.Any]


class GetAccessToken(OpenAPIAction):
    """
    The text outlines API authentication via personal tokens and OAuth flow,
    stating OAuth tokens can't be used in "Try It" feature or tested via web
    browsers.
    """

    _tags = ["Authorization"]
    _display_name = "get_access_token"
    _request_schema = GetAccessTokenRequest
    _response_schema = GetAccessTokenResponse

    url = "https://api.clickup.com/api/v2"
    path = "/oauth/token"
    method = "post"
    operation_id = "Authorization_getAccessToken"
    action_identifier = "/oauth/token_post"

    path_params = {}
    query_params = {
        "client_id": "client_id",
        "client_secret": "client_secret",
        "code": "code",
    }
    header_params = {}
    request_params = {}

    aliases = {}


class authorization_get_access_token(OpenAPIAction):
    """
    The text outlines API authentication via personal tokens and OAuth flow,
    stating OAuth tokens can't be used in "Try It" feature or tested via web
    browsers.<<DEPRECATED use get_access_token>>
    """

    _tags = ["Authorization"]
    _display_name = "authorization_get_access_token"
    _request_schema = GetAccessTokenRequest
    _response_schema = GetAccessTokenResponse

    url = "https://api.clickup.com/api/v2"
    path = "/oauth/token"
    method = "post"
    operation_id = "Authorization_getAccessToken"
    action_identifier = "/oauth/token_post"

    path_params = {}
    query_params = {
        "client_id": "client_id",
        "client_secret": "client_secret",
        "code": "code",
    }
    header_params = {}
    request_params = {}

    aliases = {}



"""Python package helpers."""

from importlib.metadata import PackageNotFoundError, metadata


_installed = set()


def check_if_package_is_intalled(name: str) -> bool:
    try:
        return name in _installed or bool(metadata(name))
    except PackageNotFoundError:
        return False


def add_package_to_installed_list(name: str):
    _installed.add(name)


# This for testing purposes
def reset_installed_list():
    _installed.clear()



import random
import string
from abc import abstractmethod
from enum import Enum
from pathlib import Path
from typing import Dict, Optional

from pydantic import BaseModel, Field

from composio.tools.base.local import ActionRequest, ActionResponse, LocalAction
from composio.tools.env.browsermanager.manager import BrowserManager


class SelectorType(str, Enum):
    CSS = "css"
    XPATH = "xpath"
    ID = "id"
    NAME = "name"
    TAG = "tag"
    CLASS = "class"


class BaseBrowserRequest(BaseModel):
    browser_manager_id: Optional[str] = Field(
        default=None,
        description="ID of the browser manager where the action will be executed. "
        "If not provided, the most recent browser manager will be used.",
    )
    capture_screenshot: bool = Field(
        default=False,
        description="Whether to capture before and after screenshots of page while executing the action.",
    )


class BaseBrowserSelectorRequest(BaseBrowserRequest):
    selector_type: str = Field(
        default="css",
        description="Type of selector to use while interacting with the element. Only `css`, `xpath`, `name`, `tag`, `class`, `id` are supported",
        examples=["css", "xpath", "name", "tag", "class", "id"],
    )
    selector: str = Field(
        ..., description="Selector value of the element to interact with"
    )
    timeout: Optional[float] = Field(
        default=None,
        description="Maximum time to wait for the action to complete (in seconds)",
    )


class BaseBrowserResponse(BaseModel):
    error: Optional[str] = Field(
        default=None,
        description="Error message if the action failed",
    )
    current_url: Optional[str] = Field(
        default=None,
        description="Current URL of the browser.",
    )
    viewport: Optional[Dict[str, Optional[int]]] = Field(
        default=None,
        description="Current viewport size of the browser window.",
    )
    scroll_position: Optional[Dict[str, Optional[int]]] = Field(
        default=None,
        description="Current scroll position of the page.",
    )
    page_dimensions: Optional[Dict[str, Optional[int]]] = Field(
        default=None,
        description="Total dimensions of the page content.",
    )
    before_screenshot: Optional[str] = Field(
        default=None,
        description="Path to the screenshot taken before the action was executed.",
    )
    after_screenshot: Optional[str] = Field(
        default=None,
        description="Path to the screenshot taken after the action was executed.",
    )


class BaseBrowserAction(LocalAction[ActionRequest, ActionResponse], abs=True):
    _tool_name: str = "browsertool"

    @abstractmethod
    def execute_on_browser_manager(
        self,
        browser_manager: BrowserManager,
        request: ActionRequest,
    ) -> BaseBrowserResponse:
        pass

    def execute(self, request: ActionRequest, metadata: dict) -> ActionResponse:
        browser_manager = self.browsers.get(request.browser_manager_id)  # type: ignore
        try:
            before_screenshot = None
            after_screenshot = None
            if request.capture_screenshot:  # type: ignore
                before_screenshot = self._take_screenshot(browser_manager, "before")

            resp = self.execute_on_browser_manager(
                browser_manager=browser_manager,
                request=request,
            )
            resp.current_url = browser_manager.get_current_url()

            # Get viewport size
            viewport = browser_manager.get_page_viewport()
            resp.viewport = (
                {k: v or None for k, v in viewport.items()} if viewport else None
            )

            # Get scroll position
            scroll_position = browser_manager.execute_script(
                """
                () => ({
                    x: window.pageXOffset,
                    y: window.pageYOffset
                })
            """
            )
            resp.scroll_position = (
                {k: v or None for k, v in scroll_position.items()}
                if scroll_position
                else None
            )

            # Get total page dimensions
            page_dimensions = browser_manager.execute_script(
                """
                () => ({
                    width: Math.max(
                        document.body.scrollWidth,
                        document.documentElement.scrollWidth,
                        document.body.offsetWidth,
                        document.documentElement.offsetWidth,
                        document.body.clientWidth,
                        document.documentElement.clientWidth
                    ),
                    height: Math.max(
                        document.body.scrollHeight,
                        document.documentElement.scrollHeight,
                        document.body.offsetHeight,
                        document.documentElement.offsetHeight,
                        document.body.clientHeight,
                        document.documentElement.clientHeight
                    )
                })
            """
            )
            resp.page_dimensions = (
                {k: v or None for k, v in page_dimensions.items()}
                if page_dimensions
                else None
            )
            if request.capture_screenshot:  # type: ignore
                after_screenshot = self._take_screenshot(browser_manager, "after")

            resp.before_screenshot = before_screenshot
            resp.after_screenshot = after_screenshot
            return resp  # type: ignore
        except Exception as e:
            error_message = (
                f"An error occurred while executing the browser action: {str(e)}"
            )
            self.logger.error(error_message, exc_info=True)
            return self.response.model(
                error=error_message,  # type: ignore
                current_url=browser_manager.get_current_url(),  # type: ignore
                viewport=None,  # type: ignore
                scroll_position=None,  # type: ignore
                page_dimensions=None,  # type: ignore
                before_screenshot=None,  # type: ignore
                after_screenshot=None,  # type: ignore
            )

    def _take_screenshot(self, browser_manager: BrowserManager, prefix: str) -> str:
        home_dir = Path.home()
        browser_media_dir = home_dir / ".browser_media"
        browser_media_dir.mkdir(parents=True, exist_ok=True)
        random_string = "".join(random.choices(string.ascii_lowercase, k=6))
        output_path = browser_media_dir / f"{prefix}_screenshot_{random_string}.png"
        browser_manager.take_screenshot(output_path, full_page=True)
        return str(output_path)



"""
Code grep tool for Composio.
"""

import typing as t

from composio.tools.base.local import LocalAction, LocalTool

from .actions import FormatAndLintCodebase


class CodeFormatTool(LocalTool, autoload=True):
    """Code Format tool."""

    logo = "https://raw.githubusercontent.com/ComposioHQ/composio/master/python/docs/imgs/logos/codeformat.png"

    @classmethod
    def actions(cls) -> t.List[t.Type[LocalAction]]:
        """Return the list of actions."""
        return [FormatAndLintCodebase]



import typing as t
from pathlib import Path

from pydantic import Field

from composio.tools.base.local import LocalAction
from composio.tools.local.filetool.actions.base_action import (
    BaseFileRequest,
    BaseFileResponse,
    include_cwd,
)


class GitCustomRequest(BaseFileRequest):
    """Request to run a custom git command."""

    cmd: str = Field(
        ...,
        description="The custom git command to run. Avoid using `git` prefix. Example: `add -u`, `commit -m 'test-commit'`",
    )


class GitCustomResponse(BaseFileResponse):
    """Response to running a custom git command."""

    success: bool = Field(
        False, description="Whether the command execution was successful"
    )
    message: str = Field("", description="Status message or error description")


class GitCustom(LocalAction[GitCustomRequest, GitCustomResponse]):
    """
    Run a custom git command from the current directory.

    This action runs a custom git command from the current working directory.

    Usage example:
    1. cmd: "add -u"
    2. cmd: "commit -m 'test-commit'"
    """

    @include_cwd  # type: ignore
    def execute(self, request: GitCustomRequest, metadata: t.Dict) -> GitCustomResponse:
        file_manager = self.filemanagers.get(request.file_manager_id)
        repo_path = Path(file_manager.working_dir)
        if not (repo_path / ".git").is_dir():
            return GitCustomResponse(
                success=False,
                message=f"Error: The directory '{repo_path}' is not the root of a git repository.",
            )

        original_dir = file_manager.working_dir
        file_manager.chdir(repo_path)
        command = f"git {request.cmd}"
        output, error = file_manager.execute_command(command)
        if error:
            return GitCustomResponse(success=False, message=f"Error: {error}")

        # Change back to the original directory
        file_manager.chdir(original_dir)
        return GitCustomResponse(
            success=True,
            message=output if output else "Git command executed successfully.",
        )



"""Virtual browser implementation using Playwright with Chromium."""

import importlib
import random
import typing as t
from pathlib import Path

from composio.utils.logging import WithLogger


selector_map = {
    "css": lambda s: s,
    "xpath": lambda s: f"xpath={s}",
    "id": lambda s: f"#{s}",
    "name": lambda s: f"[name='{s}']",
    "tag": lambda s: f"//{s}",
    "class": lambda s: f".{s}",
}


class BrowserError(Exception):
    """Exception raised for browser-related errors."""


class Browser(WithLogger):  # pylint: disable=too-many-public-methods
    """Browser object for browser manager using Chromium."""

    def __init__(
        self, headless: bool = True, window_size: t.Tuple[int, int] = (1920, 1080)
    ) -> None:
        """
        Initialize browser object

        :param headless: Whether to run browser in headless mode.
        :param window_size: Size of the browser window as (width, height).
        """
        super().__init__()

        self.headless = headless
        self.window_size = window_size
        # self.playwright: t.Optional[Playwright] = None
        # self.browser: t.Optional[PlaywrightBrowser] = None
        # self.page: t.Optional[Page] = None
        self.user_agents = [
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36",
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.1.1 Safari/605.1.15",
            "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/92.0.4515.107 Safari/537.36",
        ]
        self.current_url: str = ""
        self.playwright_module: t.Optional[t.Any] = None
        self.sync_playwright: t.Optional[t.Any] = None
        self.playwrightBrowser: t.Optional[t.Any] = None
        self.page: t.Optional[t.Any] = None
        self.playwright: t.Optional[t.Any] = None
        self.browser: t.Optional[t.Any] = None

    def load_playwright(self) -> None:
        """Load the playwright module."""
        if self.playwright_module is None:
            self.playwright_module = importlib.import_module("playwright.sync_api")
            self.sync_playwright = self.playwright_module.sync_playwright
            self.playwrightBrowser = self.playwright_module.Browser
            self.page = self.playwright_module.Page
            self.playwright = self.playwright_module.Playwright

    def setup(self) -> None:
        """Set up the Chromium browser."""
        self.load_playwright()
        if self.sync_playwright is None:
            raise BrowserError("Failed to load playwright")
        self.playwright = self.sync_playwright().start()
        user_agent = random.choice(self.user_agents)
        self.browser = self.playwright.chromium.launch(
            headless=self.headless, args=self._get_browser_args(user_agent)
        )
        self.page = self.browser.new_page(
            viewport={"width": self.window_size[0], "height": self.window_size[1]},
            user_agent=user_agent,
            locale="en-US",
            timezone_id="America/New_York",
            geolocation={"latitude": 40.7128, "longitude": -74.0060},
            permissions=["geolocation"],
        )
        self._setup_browser_environment()

        # Navigate to Google.com after setting up the browser
        self.page.goto("https://www.google.com", wait_until="networkidle")
        self.current_url = self.page.url
        self._add_random_delay()

    def _get_browser_args(self, user_agent: str) -> t.List[str]:
        """Get browser launch arguments."""
        return [
            f"--user-agent={user_agent}",
            "--disable-blink-features=AutomationControlled",
            "--no-sandbox",
            "--disable-setuid-sandbox",
            "--disable-dev-shm-usage",
            "--disable-accelerated-2d-canvas",
            "--no-first-run",
            "--no-zygote",
            "--disable-gpu",
            "--lang=en-US,en;q=0.9",
            "--disable-extensions",
            "--mute-audio",
        ]

    def _setup_browser_environment(self) -> None:
        """Set up browser environment to mimic real user."""
        if self.page:
            self.page.evaluate(
                """
                () => {
                    Object.defineProperty(navigator, 'webdriver', {
                        get: () => undefined
                    });
                    Object.defineProperty(navigator, 'languages', {
                        get: () => ['en-US', 'en']
                    });
                    Object.defineProperty(navigator, 'plugins', {
                        get: () => [1, 2, 3, 4, 5]
                    });
                }
            """
            )

    def _ensure_page_initialized(self):
        """Ensure page is initialized before operations."""
        if self.page is None:
            raise BrowserError("Page is not initialized")
        return self.page

    def goto(self, url: str, timeout: int = 60000) -> None:
        """Navigate to a specific URL."""
        page = self._ensure_page_initialized()

        # try:
        page.goto(url, wait_until="networkidle", timeout=timeout)
        # except Exception as e:
        #     self.logger.warning(
        #         f"Timeout or error occurred while waiting for networkidle: {str(e)}"
        #     )

        self.current_url = page.url
        self._add_random_delay()

    def get_page_viewport(self) -> t.Dict[str, int]:
        """Get the viewport of the current page."""
        try:
            page = self._ensure_page_initialized()
            viewport = page.viewport_size
            if viewport:
                return viewport  # type: ignore
            raise BrowserError("Failed to get page viewport")
        except BrowserError as e:
            self.logger.error(f"Failed to get page viewport: {str(e)}")
            raise

    def back(self) -> None:
        """Navigate back in browser history."""
        page = self._ensure_page_initialized()
        page.go_back()
        self.current_url = page.url
        self._add_random_delay()

    def forward(self) -> None:
        """Navigate forward in browser history."""
        page = self._ensure_page_initialized()
        page.go_forward()
        self.current_url = page.url
        self._add_random_delay()

    def refresh(self, ignore_cache: bool = False) -> None:
        """
        Reload the current page.

        :param ignore_cache: If True, the page is loaded in a new context to bypass caching.
        """
        page = self._ensure_page_initialized()
        current_url = page.url

        if ignore_cache and self.browser:
            # Create a new context and page
            context = self.browser.new_context()
            new_page = context.new_page()

            # Navigate to the current URL in the new page
            new_page.goto(current_url, wait_until="networkidle", timeout=30000)

            # Close the old page and context
            page.close()
            page.context.close()

            # Update the current page
            self.page = new_page
        else:
            # Regular reload if ignore_cache is False
            page.reload(wait_until="networkidle", timeout=30000)
        page = self._ensure_page_initialized()
        self.current_url = page.url
        self._add_random_delay()

    # def get_content(self) -> str:
    #     """Get the current page's HTML content."""
    #     page = self._ensure_page_initialized()
    #     return page.content()

    def find_element(self, selector: str, selector_type: str = "css") -> t.Any:
        """
        Find an element on the page.

        :param selector: Selector for the element.
        :param selector_type: Type of selector (CSS, XPATH, ID, NAME, TAG, CLASS). Defaults to "css".
        :return: The found element.
        """
        page = self._ensure_page_initialized()
        try:
            selector_func = selector_map.get(selector_type.lower())
            if not selector_func:
                raise ValueError(f"Unsupported selector type: {selector_type}")
            return page.query_selector(selector_func(selector))
        except Exception as e:
            raise BrowserError(
                f"Failed to find element with selector '{selector}': {str(e)}"
            ) from e

    def click(self, selector: str, selector_type: str = "css") -> None:
        """
        Click an element on the page.

        :param selector: Selector for the element to click.
        :param selector_type: Type of selector (CSS, XPATH, ID, NAME, TAG, CLASS).
        """
        page = self._ensure_page_initialized()
        try:
            selector_func = selector_map.get(selector_type.lower())
            if not selector_func:
                raise ValueError(f"Unsupported selector type: {selector_type}")
            element = page.query_selector(selector_func(selector))
            if element:
                element.click()
                self.current_url = page.url
                self._add_random_delay()
            else:
                raise BrowserError(f"Element not found with selector '{selector}'")
        except Exception as e:
            raise BrowserError(
                f"Failed to click element with selector '{selector}': {str(e)}"
            ) from e

    def type(self, selector: str, text: str, selector_type: str = "css") -> None:
        """
        Type text into an input field.

        :param selector: CSS selector for the input field.
        :param text: Text to type.
        :param selector_type: Type of selector (CSS, XPATH, ID, NAME, TAG, CLASS). Defaults to "css".
        """
        page = self._ensure_page_initialized()
        selector_func = selector_map.get(selector_type.lower())
        if not selector_func:
            raise ValueError(f"Unsupported selector type: {selector_type}")
        page.fill(selector_func(selector), text)
        self._add_random_delay()

    def clear(self, selector: str, selector_type: str = "css") -> None:
        """
        Clear the text from an input field.

        :param selector: CSS selector for the input field.
        :param selector_type: Type of selector (CSS, XPATH, ID, NAME, TAG, CLASS). Defaults to "css".
        """
        page = self._ensure_page_initialized()
        selector_func = selector_map.get(selector_type.lower())
        if not selector_func:
            raise ValueError(f"Unsupported selector type: {selector_type}")
        page.fill(selector_func(selector), "")
        self._add_random_delay()

    # def select(self, selector: str, value: str) -> None:
    #     """
    #     Select an option from a dropdown.

    #     :param selector: CSS selector for the dropdown.
    #     :param value: Value to select.
    #     """
    #     page = self._ensure_page_initialized()
    #     page.select_option(selector, value)
    #     self._add_random_delay()

    def scroll(self, direction: str, amount: int) -> None:
        """
        Scroll the page.

        :param direction: Direction to scroll ('UP', 'DOWN', 'LEFT', or 'RIGHT').
        :param amount: Number of pixels to scroll.
        """
        page = self._ensure_page_initialized()
        if direction.upper() in ["UP", "DOWN"]:
            scroll_amount = -amount if direction.upper() == "UP" else amount
            page.evaluate(f"window.scrollBy(0, {scroll_amount})")
        elif direction.upper() in ["LEFT", "RIGHT"]:
            scroll_amount = -amount if direction.upper() == "LEFT" else amount
            page.evaluate(f"window.scrollBy({scroll_amount}, 0)")
        else:
            raise ValueError(
                f"Invalid scroll direction: {direction}. Must be 'UP', 'DOWN', 'LEFT', or 'RIGHT'."
            )
        self._add_random_delay()

    def scroll_to_element(self, selector: str, selector_type: str = "css") -> None:
        """
        Scroll to a specific element if it is not already visible on the screen.

        :param selector: CSS selector for the element to scroll to.
        :param selector_type: Type of selector (CSS, XPATH, ID, NAME, TAG, CLASS). Defaults to "css".
        """
        page = self._ensure_page_initialized()
        selector_func = selector_map.get(selector_type.lower())
        if not selector_func:
            raise ValueError(f"Unsupported selector type: {selector_type}")

        # Check if the element is visible in the viewport
        is_visible = page.evaluate(
            """
            (selector) => {
                const element = document.querySelector(selector);
                if (!element) return false;
                const rect = element.getBoundingClientRect();
                const viewHeight = Math.max(document.documentElement.clientHeight, window.innerHeight);
                const viewWidth = Math.max(document.documentElement.clientWidth, window.innerWidth);
                return !(rect.bottom < 0 || rect.top - viewHeight >= 0 || rect.right < 0 || rect.left - viewWidth >= 0);
            }
        """,
            selector_func(selector),
        )

        # Only scroll to the element if it is not visible
        if not is_visible:
            page.evaluate(
                """
                (selector) => {
                    const element = document.querySelector(selector);
                    if (element) {
                        element.scrollIntoView({behavior: 'smooth', block: 'center'});
                    } else {
                        console.warn('Element not found:', selector);
                    }
                }
            """,
                selector_func(selector),
            )
            self._add_random_delay()

    def new_tab(self) -> None:
        """Open a new tab."""
        if self.browser is None:
            raise BrowserError("Browser is not initialized")
        self.page = self.browser.new_page()
        self.current_url = self.page.url

    def switch_tab(self, index: int) -> None:
        """
        Switch between tabs.

        :param index: Index of the tab to switch to.
        """
        if self.browser is None:
            raise BrowserError("Browser is not initialized")
        pages = self.browser.contexts[0].pages
        if 0 <= index < len(pages):
            self.page = pages[index]
            self.current_url = self.page.url

        raise IndexError("Tab index out of range")

    def close_tab(self) -> None:
        """Close the current tab."""
        page = self._ensure_page_initialized()
        if self.browser is None:
            raise BrowserError("Browser is not initialized")
        page.close()
        self.page = self.browser.contexts[0].pages[-1]
        self.current_url = self.page.url

    def take_screenshot(self, path: Path, full_page: bool = True) -> None:
        """
        Capture a screenshot of the current page.

        :param path: Path to save the screenshot.
        """
        page = self._ensure_page_initialized()
        page.screenshot(path=str(path), full_page=full_page)

    def press_key(self, key: str) -> None:
        """
        Press a key on the keyboard.

        :param key: Key to press.
        """
        page = self._ensure_page_initialized()
        page.keyboard.press(key)
        self._add_random_delay()

    def execute_script(self, script: str, *args) -> t.Any:
        """
        Execute a custom JavaScript script.

        :param script: JavaScript code to execute.
        :param args: Arguments to pass to the script.
        :return: Result of the script execution.
        """
        page = self._ensure_page_initialized()
        return page.evaluate(script, *args)

    def get_element_attribute(
        self, selector: str, attribute: str, selector_type: str = "css"
    ) -> t.Optional[str]:
        """
        Get the value of an attribute for a specific element.

        :param selector: Selector to find the element.
        :param attribute: Name of the attribute to get.
        :param selector_type: Type of selector (default is "css").
        :return: Value of the attribute or None if not found.
        """
        element = self.find_element(selector, selector_type)
        if element:
            return element.get_attribute(attribute)
        return None

    def get_element_text(
        self, selector: str, selector_type: str = "css"
    ) -> t.Optional[str]:
        """
        Get the text content of a specific element.

        :param selector: Selector to find the element.
        :param selector_type: Type of selector (default is "css").
        :return: Text content of the element or None if not found.
        """
        element = self.find_element(selector, selector_type)
        if element:
            return element.inner_text()
        return None

    def get_page_details(
        self,
    ) -> t.Dict[str, t.Any]:
        """Get the details of the current page."""
        page = self._ensure_page_initialized()
        details = {
            "url": page.url,
            "title": page.title(),
            "page_details": page.accessibility.snapshot(),
        }
        return details

    def __str__(self) -> str:
        """String representation."""
        return f"Browser(type=chromium, headless={self.headless}, current_url={self.current_url})"

    def cleanup(self) -> None:
        """Clean up resources."""
        if self.browser:
            self.browser.close()
        if self.playwright:
            self.playwright.stop()

    def _add_random_delay(self):
        """Add a random delay to mimic human behavior."""
        page = self._ensure_page_initialized()
        page.wait_for_timeout(random.uniform(100, 600))

    def __enter__(self):
        """Context manager entry."""
        self.setup()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit."""
        self.cleanup()



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetListViewsRequest(BaseModel):
    """Request schema for `GetListViews`"""

    list_id: int = Field(
        ...,
        alias="list_id",
        description="",
    )


class GetListViewsResponse(BaseModel):
    """Response schema for `GetListViews`"""

    data: t.Dict[str, t.Any]


class GetListViews(OpenAPIAction):
    """
    View the task and page views available for a List.<br> Views and required
    views are separate responses.
    """

    _tags = ["Views"]
    _display_name = "get_list_views"
    _request_schema = GetListViewsRequest
    _response_schema = GetListViewsResponse

    url = "https://api.clickup.com/api/v2"
    path = "/list/{list_id}/view"
    method = "get"
    operation_id = "Views_getListViews"
    action_identifier = "/list/{list_id}/view_get"

    path_params = {"list_id": "list_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



"""
FlyIO workspace implementation.
"""

import typing as t
from dataclasses import dataclass
from urllib.parse import urlparse

from composio.exceptions import ComposioSDKError
from composio.tools.env.base import RemoteWorkspace, WorkspaceConfigType
from composio.tools.env.flyio.client import (
    FLYIO_DEPENDENCIES_INSTALLED,
    FlyIO,
    PortRequest,
)


@dataclass
class Config(WorkspaceConfigType):
    """Host configuration type."""

    image: t.Optional[str] = None
    """Docker image to use for creating workspace."""

    token: t.Optional[str] = None
    """FlyIO API token."""

    ports: t.Optional[t.List[PortRequest]] = None
    """Port requests."""


class FlyIOWorkspace(RemoteWorkspace):
    """FlyIO Workspace."""

    flyio: FlyIO

    def __init__(self, config: Config):
        """Initialize FlyIO workspace."""
        if not FLYIO_DEPENDENCIES_INSTALLED:
            raise ComposioSDKError(
                "`flyio` workspace dependencies are not installed"
                "run `pip3 install composio-core[flyio]` to install."
            )

        super().__init__(config=config)
        self.image = config.image
        self.token = config.token
        self._port_requests = config.ports or []

    def setup(self) -> None:
        """Setup workspace."""
        self.flyio = FlyIO(
            access_token=self.access_token,
            image=self.image,
            flyio_token=self.token,
            environment=self.environment,
            ports=self._port_requests,
        )
        self.flyio.setup()
        self.url = self.flyio.url
        self.host = t.cast(str, urlparse(url=self.url).hostname)

        ports = []
        for r in self._port_requests:
            ports += [p["port"] for p in r["ports"]]
        self.ports = ports

    def teardown(self) -> None:
        """Teardown E2B workspace."""
        super().teardown()
        if hasattr(self, "flyio"):
            self.flyio.teardown()



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class InviteUserToWorkspaceRequest(BaseModel):
    """Request schema for `InviteUserToWorkspace`"""

    team_id: int = Field(
        ...,
        alias="team_id",
        description="Team ID (Workspace)",
    )
    email: str = Field(
        default=...,
        alias="email",
        description="Email",
    )
    admin: bool = Field(
        default=...,
        alias="admin",
        description="Admin",
    )
    custom_role_id: int = Field(
        default=None,
        alias="custom_role_id",
        description="Custom Role Id",
    )


class InviteUserToWorkspaceResponse(BaseModel):
    """Response schema for `InviteUserToWorkspace`"""

    data: t.Dict[str, t.Any]


class InviteUserToWorkspace(OpenAPIAction):
    """
    Invite someone to your Workspace as a member through a specific endpoint.
    Note: Guest invitations and this feature are exclusive to Workspaces on
    the Enterprise Plan.
    """

    _tags = ["Users"]
    _display_name = "invite_user_to_workspace"
    _request_schema = InviteUserToWorkspaceRequest
    _response_schema = InviteUserToWorkspaceResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/user"
    method = "post"
    operation_id = "Users_inviteUserToWorkspace"
    action_identifier = "/team/{team_id}/user_post"

    path_params = {"team_id": "team_id"}
    query_params = {}
    header_params = {}
    request_params = {
        "email": {"__alias": "email"},
        "admin": {"__alias": "admin"},
        "custom_role_id": {"__alias": "custom_role_id"},
    }

    aliases = {}



"""
Action for getting a screenshot of a webpage.
"""

import random
import string
from pathlib import Path

from pydantic import Field

from composio.tools.env.browsermanager.manager import BrowserManager
from composio.tools.local.browsertool.actions.base_action import (
    BaseBrowserAction,
    BaseBrowserRequest,
    BaseBrowserResponse,
)


class GetScreenshotRequest(BaseBrowserRequest):
    """Request schema for getting a screenshot."""

    output_path: str = Field(
        default="",
        description="""Optional path to save the screenshot. Preferable is not provide this.
                            If not provided, a default path will be used.
                            Example: '/home/user/screenshot.png'""",
    )
    full_page: bool = Field(
        default=True,
        description="Whether to take a full page screenshot or just the visible area.",
    )


class GetScreenshotResponse(BaseBrowserResponse):
    """Response schema for getting a screenshot."""

    screenshot_path: str = Field(
        default="", description="Path where the screenshot was saved"
    )
    success: bool = Field(
        default=False, description="Whether the screenshot action was successful"
    )


class GetScreenshot(BaseBrowserAction[GetScreenshotRequest, GetScreenshotResponse]):
    """
    Get a screenshot of a webpage.

    If a URL is provided, the page will be newly loaded before taking the screenshot.
    If no URL is provided, the screenshot will be taken of the current open page.
    This action allows capturing a screenshot of a specified webpage or the current page
    in the browser. It can take full-page screenshots or just the visible area.

    If no output path is specified, it generates a random filename in the user's
    .browser_media directory.
    """

    display_name = "GetScreenshot"
    _request_schema = GetScreenshotRequest
    _response_schema = GetScreenshotResponse

    def execute_on_browser_manager(
        self, browser_manager: BrowserManager, request: GetScreenshotRequest  # type: ignore
    ) -> GetScreenshotResponse:
        """Execute the screenshot action."""
        try:
            if not request.output_path or request.output_path == "":
                home_dir = Path.home()
                browser_media_dir = home_dir / ".browser_media"
                browser_media_dir.mkdir(parents=True, exist_ok=True)
                random_string = "".join(random.choices(string.ascii_lowercase, k=6))
                output_path = browser_media_dir / f"screenshot_{random_string}.png"
            else:
                output_path = Path(request.output_path)
            browser_manager.take_screenshot(output_path, full_page=request.full_page)
            return GetScreenshotResponse(screenshot_path=str(output_path), success=True)
        except Exception as e:
            return GetScreenshotResponse(
                screenshot_path="",
                success=False,
                error=f"Unexpected error while taking screenshot: {str(e)}",
            )



import os
import threading
import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from uuid import uuid4

import requests

from composio.client.enums import Action, ActionType, AppType, TagType
from composio.constants import ENV_COMPOSIO_API_KEY, ENV_COMPOSIO_BASE_URL
from composio.exceptions import ComposioSDKError
from composio.tools.env.id import generate_id
from composio.utils.logging import WithLogger


ENV_GITHUB_ACCESS_TOKEN = "GITHUB_ACCESS_TOKEN"
ENV_ACCESS_TOKEN = "ACCESS_TOKEN"

WORKSPACE_PROMPT = """You have access to a workspace with open {ports} network
ports being available publicly and hostname to reach this machine is {host}, 
you can use this for development and deployment purposes.
"""


def _read_env_var(name: str, default: t.Any) -> str:
    """Read environment variable."""
    if default is not None:
        return default

    value = os.environ.get(name, default)
    if value is None:
        raise ComposioSDKError(f"Please provide value for `{name}`")
    return value


class Sessionable(WithLogger, ABC):
    """Sessionable abstraction"""

    _id: str

    def __str__(self) -> str:
        """String representation."""
        return f"Sessionable(type={self.__class__.__name__}, id={self.id})"

    __repr__ = __str__

    @property
    def id(self) -> str:
        """Get shell ID."""
        return self._id

    @abstractmethod
    def setup(self) -> None:
        """Setup session."""

    @abstractmethod
    def teardown(self) -> None:
        """Teardown session."""


SessionableType = t.TypeVar("SessionableType", bound=Sessionable)


class SessionFactory(WithLogger, t.Generic[SessionableType]):
    """Factory abstraction."""

    _session: t.Dict[str, SessionableType] = {}
    _recent: t.Optional[SessionableType] = None
    _lock: threading.Lock = threading.Lock()

    def __init__(self, factory: t.Callable[[], SessionableType]) -> None:
        """Creatte shell factory"""
        super().__init__()
        self._factory = factory

    @property
    def recent(self) -> SessionableType:
        """Get most recent workspace."""
        with self._lock:
            session = self._recent
        if session is None:
            session = self.new()
            with self._lock:
                self._recent = session
        return session

    @recent.setter
    def recent(self, shell: SessionableType) -> None:
        """Get most recent workspace."""
        with self._lock:
            self._recent = shell

    def new(self) -> SessionableType:
        """Create a new shell."""
        session = self._factory()
        session.setup()
        self._session[session.id] = session
        self.recent = session
        return session

    def get(self, id: t.Optional[str] = None) -> SessionableType:
        """Get shell instance."""
        if id is None or id == "":
            return self.recent

        if id not in self._session:
            raise ComposioSDKError(
                message=f"No session of type {self._factory.__name__} found with ID: {id}",
            )
        shell = self._session[id]
        self.recent = shell
        return shell

    def teardown(self) -> None:
        """Stop all running shells."""
        while len(self._session) > 0:
            id = list(self._session.keys()).pop()
            session = self._session.pop(id)
            session.teardown()
        self._recent = None


@dataclass
class WorkspaceConfigType:
    """Workspace configuration."""

    composio_api_key: t.Optional[str] = None
    """Composio API Key."""

    composio_base_url: t.Optional[str] = None
    """Base URL for composio backend."""

    github_access_token: t.Optional[str] = None
    """Github access token agent workspace, if not provided the access token from the active composio account will be used."""

    environment: t.Optional[t.Dict[str, str]] = None
    """Environment config for workspace."""

    persistent: bool = False
    """Set `True` to make this workspace persistent."""


class Workspace(WithLogger, ABC):
    """Workspace abstraction for executing tools."""

    url: str
    """URL for the tooling server (Only applicable for remote workspace)."""

    host: str
    """Host string for the workspace."""

    ports: t.List[int]
    """List of available ports on the workspace, if empty all of the ports are available."""

    def __init__(self, config: WorkspaceConfigType):
        """Initialize workspace."""
        super().__init__()
        self.id = generate_id()
        self.access_token = uuid4().hex.replace("-", "")
        self.composio_api_key = _read_env_var(
            name=ENV_COMPOSIO_API_KEY,
            default=config.composio_api_key,
        )
        self.composio_base_url = _read_env_var(
            name=ENV_COMPOSIO_BASE_URL,
            default=config.composio_base_url,
        )
        self.github_access_token = config.github_access_token or os.environ.get(
            ENV_GITHUB_ACCESS_TOKEN, "NO_VALUE"
        )
        self.environment = {
            **(config.environment or {}),
            ENV_COMPOSIO_API_KEY: self.composio_api_key,
            ENV_COMPOSIO_BASE_URL: self.composio_base_url,
            ENV_GITHUB_ACCESS_TOKEN: self.github_access_token,
            f"_COMPOSIO_{ENV_GITHUB_ACCESS_TOKEN}": self.github_access_token,
            ENV_ACCESS_TOKEN: self.access_token,
        }
        self.persistent = config.persistent

    def __str__(self) -> str:
        """String representation."""
        return f"Workspace(type={self.__class__.__name__}, id={self.id})"

    __repr__ = __str__

    def as_prompt(self) -> str:
        """Format current workspace details for the agentic use."""
        return WORKSPACE_PROMPT.format(ports=self.ports, host=self.host)

    @abstractmethod
    def check_for_missing_dependencies(
        self,
        apps: t.Optional[t.Sequence[AppType]] = None,
        actions: t.Optional[t.Sequence[ActionType]] = None,
        tags: t.Optional[t.Sequence[TagType]] = None,
    ) -> None:
        """Install dependecies in the given workspace."""

    @abstractmethod
    def setup(self) -> None:
        """Setup workspace."""

    @abstractmethod
    def execute_action(
        self,
        action: Action,
        request_data: dict,
        metadata: dict,
    ) -> t.Dict:
        """Execute an action in this workspace."""

    @abstractmethod
    def teardown(self) -> None:
        """Teardown current workspace."""


class RemoteWorkspace(Workspace):
    """Remote workspace client."""

    def _request(
        self,
        endpoint: str,
        method: str,
        json: t.Optional[t.Dict] = None,
        timeout: t.Optional[float] = 300.0,
        log: bool = True,
    ) -> requests.Response:
        """Make request to the tooling server."""
        response = requests.request(
            url=f"{self.url}{endpoint}",
            method=method,
            json=json,
            headers={
                "x-api-key": self.access_token,
            },
            timeout=timeout,
        )
        if log:
            self.logger.debug(
                f"Making HTTP request on {self.id}\n"
                f"Request: {method.upper()} {endpoint} @ {self.url}\n"
                f"Response: {response.status_code} -> {response.text}"
            )
        if response.status_code in (500, 503):
            raise ComposioSDKError(
                message=(
                    f"Error requesting data from {self}, "
                    f"Request: {method.upper()} {endpoint} @ {self.url}\n"
                    f"Response: {response.status_code} -> {response.text}"
                ),
            )
        return response

    def _upload(self, action: Action) -> None:
        """Upload action instance to tooling server."""
        from composio.tools.base.abs import (  # pylint: disable=import-outside-toplevel
            tool_registry,
        )

        obj = tool_registry["runtime"][action.app].get(action)
        request = self._request(
            method="post",
            endpoint="/tools",
            json={
                "content": Path(obj.file).read_text(encoding="utf-8"),
                "filename": Path(obj.file).name,
                "dependencies": obj.requires or [],
            },
        )
        response = request.json()
        if response["error"] is not None:
            self.logger.error(
                f"Error while uploading {action.slug}: " + response["error"]
            )
            return

        self.logger.debug(
            f"Successfully uploaded: {action.slug} - {response}",
        )
        return

    def check_for_missing_dependencies(
        self,
        apps: t.Optional[t.Sequence[AppType]] = None,
        actions: t.Optional[t.Sequence[ActionType]] = None,
        tags: t.Optional[t.Sequence[TagType]] = None,
    ) -> None:
        request = self._request(
            endpoint="/validate",
            method="post",
            json={
                "apps": list(map(str, apps or [])),
                "actions": list(
                    map(str, filter(lambda x: not hasattr(x, "enum"), actions or []))
                ),
                "tags": list(map(str, tags or [])),
            },
            timeout=600,
        )
        if request.status_code != 200:
            raise ComposioSDKError(f"Error installing dependencies: {request.text}")

        response = request.json()
        if response["error"] is not None:
            raise ComposioSDKError(
                f"Error installing dependencies: {response['error']}"
            )

    def execute_action(
        self,
        action: Action,
        request_data: dict,
        metadata: dict,
    ) -> t.Dict:
        """Execute action in docker workspace."""
        if action.is_runtime:
            self._upload(action=action)

        _ = metadata.pop("_toolset", None)
        request = self._request(
            method="post",
            endpoint=f"/actions/execute/{action.slug}",
            json={
                "params": request_data,
                "metadata": metadata,
            },
        )
        response = request.json()
        if response["error"] is None:
            return response["data"]
        raise RuntimeError(f"Error while executing {action.slug}: " + response["error"])



import typing as t

from pydantic import Field

from composio.tools.base.exceptions import ExecutionFailed
from composio.tools.base.local import LocalAction
from composio.tools.local.filetool.actions.base_action import (
    BaseFileRequest,
    BaseFileResponse,
    include_cwd,
)


class OpenFileRequest(BaseFileRequest):
    """Request to open a file."""

    file_path: str = Field(
        ...,
        description="file path to open in the editor. This is a relative path to the current directory",
    )
    line_number: int = Field(
        default=0,
        description="If file-number is given, file will be opened from that line number",
    )


class OpenFileResponse(BaseFileResponse):
    """Response to open a file."""

    message: str = Field(default="", description="Message to display to the user")
    lines: str = Field(default="", description="File content with their line numbers")


class OpenFile(LocalAction[OpenFileRequest, OpenFileResponse]):
    """
    Opens a file in the editor based on the provided file path,
    If line_number is provided, the window will be moved after that line. (i.e. 100 lines after the line number will be displayed)

    Can result in:
    - ValueError: If file_path is not a string or if the file does not exist.
    - FileNotFoundError: If the file does not exist.
    - IOError: If there's an issue reading the file.
    - PermissionError: If the user doesn't have permission to read the file.
    - IsADirectoryError: If the provided path is a directory.
    """

    @include_cwd  # type: ignore
    def execute(self, request: OpenFileRequest, metadata: t.Dict) -> OpenFileResponse:
        """Open a file."""
        try:
            file = self.filemanagers.get(request.file_manager_id).open(
                request.file_path
            )
            if request.line_number > 0:
                file.goto(request.line_number)

            content = file.format_text(lines=file.read())
            if len(content) == 0:
                raise ExecutionFailed("File is empty")

            return OpenFileResponse(
                message="File opened successfully. 100 lines after the cursor displayed.",
                lines=content,
            )
        except FileNotFoundError as e:
            raise ExecutionFailed(f"File not found: {str(e)}") from e
        except IsADirectoryError as e:
            raise ExecutionFailed(f"Cannot open a directory: {str(e)}") from e
        except PermissionError as e:
            raise ExecutionFailed(f"Permission denied: {str(e)}") from e
        except IOError as e:
            raise ExecutionFailed(f"Error reading file: {str(e)}") from e



from datetime import datetime
from typing import Dict

from pydantic import BaseModel, Field

from composio.constants import LOCAL_CACHE_DIRECTORY
from composio.tools.base.local import LocalAction


class ScreenCaptureRequest(BaseModel):
    pass


class ScreenCaptureResponse(BaseModel):
    file_path: str = Field(..., description="Path where the screenshot was saved")


class ScreenCapture(LocalAction[ScreenCaptureRequest, ScreenCaptureResponse]):
    """
    Useful to capture a screenshot of the current screen.
    """

    display_name = "Capture a screenshot"
    requires = ["pyautogui"]

    def execute(
        self,
        request: ScreenCaptureRequest,
        metadata: Dict,
    ) -> ScreenCaptureResponse:
        # pylint: disable=import-outside-toplevel
        import pyautogui

        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        file_path = LOCAL_CACHE_DIRECTORY / "output" / f"screenshot_{timestamp}.png"
        screenshot = pyautogui.screenshot()
        screenshot.save(file_path)
        return ScreenCaptureResponse(file_path=str(file_path))



"""
Apps manager for Composio SDK.

Usage:
    composio apps [command] [options]
"""

import ast
import inspect
import typing as t
from pathlib import Path

import click

from composio.cli.context import Context, get_context, pass_context
from composio.cli.utils.decorators import handle_exceptions
from composio.cli.utils.helpfulcmd import HelpfulCmdBase
from composio.client import Composio, enums
from composio.client.collections import ActionModel, AppModel, TriggerModel
from composio.core.cls.did_you_mean import DYMGroup
from composio.tools.base.abs import DEPRECATED_MARKER
from composio.tools.local import load_local_tools
from composio.utils import get_enum_key


class AppsExamples(HelpfulCmdBase, DYMGroup):
    examples = [
        click.style("composio apps", fg="green")
        + click.style("            # List all apps\n", fg="black"),
        click.style("composio apps --enabled", fg="green")
        + click.style("  # List only enabled apps\n", fg="black"),
        click.style("composio apps update", fg="green")
        + click.style("     # Update local Apps database\n", fg="black"),
    ]


@click.group(name="apps", invoke_without_command=True, cls=AppsExamples)
@click.help_option("--help", "-h", "-help")
@click.option(
    "--enabled",
    is_flag=True,
    default=False,
    help="Only show apps which are enabled",
)
@handle_exceptions()
@pass_context
def _apps(context: Context, enabled: bool = False) -> None:
    """List composio tools/apps which you have access to"""
    if context.click_ctx.invoked_subcommand:
        return

    apps = context.client.apps.get()
    if enabled:
        apps = [app for app in apps if app.enabled]
        context.console.print("[green]Showing apps which are enabled[/green]")
    else:
        context.console.print("[green]Showing all apps[/green]")

    for app in apps:
        context.console.print(f"• {app.key}")


class UpdateExamples(HelpfulCmdBase, click.Command):
    examples = [
        click.style("composio apps update", fg="green")
        + click.style("  # Update local Apps database\n", fg="black"),
    ]


@_apps.command(name="update", cls=UpdateExamples)
@click.option(
    "--beta",
    is_flag=True,
    help="Include beta apps.",
)
@click.help_option("--help", "-h", "-help")
@handle_exceptions()
@pass_context
def _update(context: Context, beta: bool = False) -> None:
    """Updates local Apps database."""
    apps = update_apps(client=context.client, beta=beta)
    update_actions(client=context.client, apps=apps, beta=beta)
    update_triggers(client=context.client, apps=apps, beta=beta)


def filter_non_beta_items(items):
    filtered_items = []
    for item in items:
        if not item.name.lower().endswith("beta"):
            filtered_items.append(item)

    seen = set()
    unique_items = []
    for item in filtered_items:
        if item.name not in seen:
            unique_items.append(item)
            seen.add(item.name)
    return unique_items


def update_apps(client: Composio, beta: bool = False) -> t.List[AppModel]:
    """Update apps."""
    apps = sorted(
        client.apps.get(),
        key=lambda x: x.key,
    )
    if not beta:
        apps = filter_non_beta_items(apps)

    _update_apps(apps=apps)
    return apps


def update_actions(
    client: Composio, apps: t.List[AppModel], beta: bool = False
) -> None:
    """Update actions and tags."""
    actions = sorted(
        client.actions.get(allow_all=True),
        key=lambda x: f"{x.appName}_{x.name}",
    )
    if not beta:
        actions = filter_non_beta_items(actions)

    _update_tags(apps=apps, actions=actions)
    _update_actions(apps=apps, actions=actions)


def update_triggers(
    client: Composio, apps: t.List[AppModel], beta: bool = False
) -> None:
    """Update triggers."""
    triggers = sorted(
        client.triggers.get(),
        key=lambda x: f"{x.appKey}_{x.name}",
    )
    if not beta:
        triggers = filter_non_beta_items(triggers)

    _update_triggers(apps=apps, triggers=triggers)


def _update_apps(apps: t.List[AppModel]) -> None:
    """Create App enum class."""
    app_names = []
    enums.base.APPS_CACHE.mkdir(
        parents=True,
        exist_ok=True,
    )
    for app in apps:
        app_names.append(
            get_enum_key(
                name=app.key.lower().replace(" ", "_").replace("-", "_"),
            )
        )
        enums.base.AppData(
            name=app.name,
            path=enums.base.APPS_CACHE / app_names[-1],
            is_local=False,
        ).store()

    for tool in load_local_tools()["local"].values():
        if tool.enum in app_names:
            continue

        app_names.append(tool.enum)
        enums.base.AppData(
            name=tool.name,
            path=enums.base.APPS_CACHE / app_names[-1],
            is_local=True,
        ).store()

    _update_annotations(
        cls=enums.App,
        attributes=app_names,
    )


def _update_actions(apps: t.List[AppModel], actions: t.List[ActionModel]) -> None:
    """Get Action enum."""
    enums.base.ACTIONS_CACHE.mkdir(parents=True, exist_ok=True)
    deprecated = {}
    action_names = []
    for app in sorted(apps, key=lambda x: x.key):
        for action in actions:
            if action.appName != app.key:
                continue

            if (
                action.description is not None
                and DEPRECATED_MARKER in action.description
            ):
                _, newact = action.description.split(DEPRECATED_MARKER, maxsplit=1)
                deprecated[get_enum_key(name=action.name)] = (
                    action.appName.lower() + "_" + newact.replace(">>", "")
                ).upper()
            else:
                action_names.append(get_enum_key(name=action.name))

            enums.base.ActionData(
                name=action.name,
                app=app.key,
                tags=action.tags,
                no_auth=app.no_auth,
                is_local=False,
                path=enums.base.ACTIONS_CACHE / get_enum_key(name=action.name),
            ).store()

    processed = []
    for tool in load_local_tools()["local"].values():
        if tool.name in processed:
            continue

        processed.append(tool.name)
        for actcls in tool.actions():
            action_names.append(actcls.enum)
            enums.base.ActionData(
                name=actcls.enum,
                app=tool.name,
                tags=actcls.tags(),
                no_auth=True,
                is_local=True,
                path=enums.base.ACTIONS_CACHE / action_names[-1],
                shell=False,
            ).store()

    _update_annotations(
        cls=enums.Action,
        attributes=action_names,
        deprecated=deprecated,
    )


def _update_tags(apps: t.List[AppModel], actions: t.List[ActionModel]) -> None:
    """Create Tag enum class."""
    enums.base.TAGS_CACHE.mkdir(parents=True, exist_ok=True)
    tag_map: t.Dict[str, t.Set[str]] = {}
    for app in apps:
        app_name = app.key
        for action in [action for action in actions if action.appName == app_name]:
            if app_name not in tag_map:
                tag_map[app_name] = set()
            tag_map[app_name].update(action.tags or [])

    tag_names = ["DEFAULT"]
    for app_name in sorted(tag_map):
        for tag in sorted(tag_map[app_name]):
            tag_name = get_enum_key(name=f"{app_name}_{tag}")
            tag_names.append(tag_name)
            enums.base.TagData(
                app=app_name,
                value=tag,
                path=enums.base.TAGS_CACHE / tag_names[-1],
            ).store()

    enums.base.TagData(
        app="default",
        value="important",
        path=enums.base.TAGS_CACHE / "DEFAULT",
    )
    _update_annotations(
        cls=enums.Tag,
        attributes=tag_names,
    )


def _update_triggers(
    apps: t.List[AppModel],
    triggers: t.List[TriggerModel],
) -> None:
    """Get Trigger enum."""
    trigger_names = []
    enums.base.TRIGGERS_CACHE.mkdir(exist_ok=True)
    for app in apps:
        for trigger in triggers:
            if trigger.appKey != app.key:
                continue

            trigger_names.append(get_enum_key(name=trigger.name).upper())
            enums.base.TriggerData(
                name=trigger.name,
                app=app.key,
                path=enums.base.TRIGGERS_CACHE / trigger_names[-1],
            ).store()

    _update_annotations(
        cls=enums.Trigger,
        attributes=trigger_names,
    )


def _update_annotations(
    cls: t.Type,
    attributes: t.List[str],
    deprecated: t.Optional[t.Dict[str, str]] = None,
) -> None:
    """Update annontations for `cls`"""
    console = get_context().console
    file = Path(inspect.getmodule(cls).__file__)  # type: ignore
    annotations = []
    for attribute in sorted(attributes):
        annotations.append(
            ast.AnnAssign(
                target=ast.Name(id=attribute),
                annotation=ast.Constant(value=f"{cls.__name__}"),
                simple=1,
            ),
        )

    _deprecated = []
    _deprecated_names = []
    deprecated = deprecated or {}
    for old, new in deprecated.items():
        if old.lower() == new.lower():
            continue

        if new.upper() not in attributes:
            continue

        _deprecated.append(_build_deprecated_node(old=old, new=new))
        _deprecated_names.append(old.upper())

    tree = ast.parse(file.read_text(encoding="utf-8"))
    for node in tree.body:
        if not isinstance(node, ast.ClassDef):
            continue
        if node.name != cls.__name__:
            continue

        cls_attributes = [
            child.target.id  # type: ignore
            for child in node.body[1:]
            if isinstance(child, ast.AnnAssign)
            and child.target.id != "_deprecated"  # type: ignore
        ]
        if set(cls_attributes) == set(attributes):
            console.print(f"[yellow]⚠️ {cls.__name__}s does not require update[/yellow]")
            return

        def _filter(child: ast.AST) -> bool:
            if isinstance(child, ast.AnnAssign) and child.target.id == "_deprecated":  # type: ignore
                child.value = ast.Dict(
                    keys=list(map(ast.Constant, deprecated.keys())),  # type: ignore
                    values=list(map(ast.Constant, deprecated.values())),  # type: ignore
                )
                return True
            if isinstance(child, ast.AnnAssign):
                return False
            if isinstance(child, ast.FunctionDef) and child.name in _deprecated_names:
                return False
            if "@te.deprecated" in ast.unparse(child):
                return False
            return True

        body = [child for child in node.body[1:] if _filter(child=child)]
        node.body = node.body[:1] + annotations + _deprecated + body
        break

    code = ast.unparse(tree)
    code = code.replace(
        "@classmethod",
        "@classmethod  # type: ignore",
    )
    code = code.replace(
        "import typing as t",
        "\n# pylint: disable=too-many-public-methods, unused-import\n\nimport typing as t"
        "\nimport typing_extensions as te  # noqa: F401",
    )
    with file.open("w", encoding="utf-8") as fp:
        fp.write(code)
    console.print(f"[green]✔ {cls.__name__}s updated[/green]")


def _build_deprecated_node(old: str, new: str) -> ast.FunctionDef:
    """Function definition."""
    return ast.FunctionDef(
        name=old.upper(),
        args=ast.arguments(
            posonlyargs=[],
            args=[ast.arg(arg="cls")],
            kwonlyargs=[],
            kw_defaults=[],
            defaults=[],
        ),
        body=[
            ast.Return(
                value=ast.Attribute(
                    value=ast.Name(id="cls", ctx=ast.Load()),
                    attr=new.upper(),
                    ctx=ast.Load(),
                )
            )
        ],
        decorator_list=[
            ast.Name(id="classmethod", ctx=ast.Load()),
            ast.Name(id="property", ctx=ast.Load()),
            ast.Call(
                func=ast.Attribute(
                    value=ast.Name(id="te", ctx=ast.Load()),
                    attr="deprecated",
                    ctx=ast.Load(),
                ),
                args=[ast.Constant(value=f"Use {new.upper()} instead.")],
                keywords=[],
            ),
        ],
        returns=ast.Constant(value="Action"),
        lineno=0,
    )



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class DeleteListRequest(BaseModel):
    """Request schema for `DeleteList`"""

    list_id: int = Field(
        ...,
        alias="list_id",
        description="",
    )


class DeleteListResponse(BaseModel):
    """Response schema for `DeleteList`"""

    data: t.Dict[str, t.Any]


class DeleteList(OpenAPIAction):
    """Delete a List from your Workspace."""

    _tags = ["Lists"]
    _display_name = "delete_list"
    _request_schema = DeleteListRequest
    _response_schema = DeleteListResponse

    url = "https://api.clickup.com/api/v2"
    path = "/list/{list_id}"
    method = "delete"
    operation_id = "Lists_removeList"
    action_identifier = "/list/{list_id}_delete"

    path_params = {"list_id": "list_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class EditChecklistItemRequest(BaseModel):
    """Request schema for `EditChecklistItem`"""

    checklist_id: str = Field(
        ...,
        alias="checklist_id",
        description="b8a8-48d8-a0c6-b4200788a683 (uuid)",
    )
    checklist_item_id: str = Field(
        ...,
        alias="checklist_item_id",
        description="e491-47f5-9fd8-d1dc4cedcc6f (uuid)",
    )
    name: str = Field(
        default=None,
        alias="name",
        description="Name",
    )
    assignee: str = Field(
        default=None,
        alias="assignee",
        description="Assignee",
    )
    resolved: bool = Field(
        default=None,
        alias="resolved",
        description="Resolved",
    )
    parent: str = Field(
        default=None,
        alias="parent",
        description=(
            "To nest a checklist item under another checklist item, include the other "
            'item"s `checklist_item_id`. '
        ),
    )


class EditChecklistItemResponse(BaseModel):
    """Response schema for `EditChecklistItem`"""

    data: t.Dict[str, t.Any]


class EditChecklistItem(OpenAPIAction):
    """
    Update an individual line item in a task checklist.    You can rename it,
    set the assignee, mark it as resolved, or nest it under another checklist
    item.
    """

    _tags = ["Task Checklists"]
    _display_name = "edit_checklist_item"
    _request_schema = EditChecklistItemRequest
    _response_schema = EditChecklistItemResponse

    url = "https://api.clickup.com/api/v2"
    path = "/checklist/{checklist_id}/checklist_item/{checklist_item_id}"
    method = "put"
    operation_id = "TaskChecklists_updateChecklistItem"
    action_identifier = (
        "/checklist/{checklist_id}/checklist_item/{checklist_item_id}_put"
    )

    path_params = {
        "checklist_id": "checklist_id",
        "checklist_item_id": "checklist_item_id",
    }
    query_params = {}
    header_params = {}
    request_params = {
        "name": {"__alias": "name"},
        "assignee": {"__alias": "assignee"},
        "resolved": {"__alias": "resolved"},
        "parent": {"__alias": "parent"},
    }

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class DeleteViewRequest(BaseModel):
    """Request schema for `DeleteView`"""

    view_id: str = Field(
        ...,
        alias="view_id",
        description="105 (string)",
    )


class DeleteViewResponse(BaseModel):
    """Response schema for `DeleteView`"""

    data: t.Dict[str, t.Any]


class DeleteView(OpenAPIAction):
    """
    Deletes a specific View by its ID. Requires 'view_id' as a string parameter
    in the path. Successful deletion returns 200 with an empty JSON object.
    """

    _tags = ["Views"]
    _display_name = "delete_view"
    _request_schema = DeleteViewRequest
    _response_schema = DeleteViewResponse

    url = "https://api.clickup.com/api/v2"
    path = "/view/{view_id}"
    method = "delete"
    operation_id = "Views_deleteViewById"
    action_identifier = "/view/{view_id}_delete"

    path_params = {"view_id": "view_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



"""
Setup configuration for compsio core.
"""

import typing as t
from pathlib import Path

from setuptools import find_packages, setup

COMPOSIO = Path(__file__).parent.resolve() / "composio"


def scan_for_package_data(
    directory: Path,
    package: Path,
    data: t.Optional[t.List[str]] = None,
) -> t.List[str]:
    """Walk the package and scan for package files."""
    data = data or []
    for child in directory.iterdir():
        if child.name.endswith(".py") or child.name.endswith(".pyc"):
            continue

        if child.is_file():
            data.append(str(child.relative_to(package)))
            continue

        data += scan_for_package_data(
            directory=child,
            package=package,
        )
    return data


core_requirements = [
    "aiohttp",
    "requests>=2.31.0,<3",
    "jsonschema>=4.21.1,<5",
    "sentry-sdk>=2.0.0",
    "pysher==1.0.8",
    "pydantic>=2.6.4,<2.10",
    "importlib-metadata>=4.8.1",
    "jsonref>=1.1.0",
    "inflection>=0.5.1",
    "semver>=2.13.0",
    # CLI dependencies
    "click",
    "rich>=13.7.1,<14",
    "pyperclip>=1.8.2,<2",
    # Workspace dependencies
    "paramiko>=3.4.1",  # Host workspace
    # Tooling server dependencies
    "fastapi",
    "uvicorn",
]

e2b_workspace_requirements = [
    "e2b>=0.17.2a37,<1",  # E2B Workspace
    "e2b-code-interpreter",  # E2B workspace
]

docker_workspace_requirements = [
    "docker>=7.1.0",  # Docker workspace
]

flyio_workspace_requirements = [
    "gql",  # FlyIO workspace
    "requests_toolbelt",  # FlyIO workspace
]

tools_requirements = [
    "tree_sitter_languages",
    "tree_sitter==0.21.3",
    "pygments",
    "pathspec",
    "diskcache",
    "networkx",
    "ruff",
    "flake8",
    "transformers",
]

all_requirements = (
    core_requirements
    + e2b_workspace_requirements
    + docker_workspace_requirements
    + flyio_workspace_requirements
    + tools_requirements
)

setup(
    name="composio_core",
    version="0.5.48-rc.1",
    author="Utkarsh",
    author_email="utkarsh@composio.dev",
    description="Core package to act as a bridge between composio platform and other services.",
    long_description=(Path(__file__).parent / "README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    url="https://github.com/composiohq/composio",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9,<4",
    packages=find_packages(include=["composio*"]),
    entry_points={
        "console_scripts": [
            "composio=composio.cli:composio",
        ],
    },
    install_requires=core_requirements,
    extras_require={
        "all": all_requirements,
        "tools": tools_requirements,
        "e2b": e2b_workspace_requirements,
        "flyio": flyio_workspace_requirements,
        "docker": docker_workspace_requirements,
    },
    include_package_data=True,
    package_data={
        "composio": scan_for_package_data(
            directory=COMPOSIO,
            package=COMPOSIO,
        ),
    },
)



"""
Browser tool for Composio.
"""

import typing as t

from composio.tools.base.local import LocalAction, LocalTool

from .actions import (
    ClickElement,
    GetElementDetails,
    GetPageDetails,
    GetScreenshot,
    GotoPage,
    NavigateHistory,
    RefreshPage,
    ScrollPage,
    TypeText,
)


class BrowserTool(LocalTool, autoload=True):
    """Browser tool for local usage."""

    requires = ["playwright"]

    logo = "https://raw.githubusercontent.com/ComposioHQ/composio/master/python/docs/imgs/logos/browsertool.png"

    @classmethod
    def actions(cls) -> t.List[t.Type[LocalAction]]:
        """Return the list of actions."""
        return [
            GetScreenshot,
            GetPageDetails,
            ClickElement,
            GetElementDetails,
            NavigateHistory,
            RefreshPage,
            ScrollPage,
            TypeText,
            GotoPage,
        ]



"""
User information.

Usage:
    composio whoami
"""

import click

from composio.cli.context import Context, pass_context
from composio.cli.utils.decorators import handle_exceptions
from composio.cli.utils.helpfulcmd import HelpfulCmdBase


class WhoamiExamples(HelpfulCmdBase, click.Command):
    examples = [
        click.style("composio whoami", fg="green")
        + click.style("  # Display your account information\n", fg="black"),
    ]


@click.command(name="whoami", cls=WhoamiExamples)
@click.help_option("--help", "-h", "-help")
@handle_exceptions()
@pass_context
def _whoami(context: Context) -> None:
    """List your account information"""
    context.console.print(f"API Key: [green]{context.user_data.api_key}[/green]")



"""Constants for this module."""

EXIT_CODE = "exit_code"
STDOUT = "stdout"
STDERR = "stderr"

ECHO_EXIT_CODE = "echo $?"
DEFAULT_IMAGE = "composio/composio"

ENV_COMPOSIO_DEV_MODE = "COMPOSIO_DEV_MODE"
ENV_COMPOSIO_TOOLSERVER_IMAGE = "COMPOSIO_TOOLSERVER_IMAGE"



from typing import Dict

from pydantic import Field

from composio.tools.base.local import LocalAction
from composio.tools.local.filetool.actions.base_action import (
    BaseFileRequest,
    BaseFileResponse,
    include_cwd,
)


class WriteRequest(BaseFileRequest):
    """Request to write a file."""

    file_path: str = Field(
        default=None,
        description=(
            "The path to the file that will be edited. If not provided, "
            "THE CURRENTLY OPEN FILE will be edited. If provided, the "
            "file at the provided path will be OPENED and edited, changing "
            "the opened file."
        ),
    )
    text: str = Field(
        ...,
        description="The text that will replace the specified line range in the file.",
    )


class WriteResponse(BaseFileResponse):
    """Response to write a file."""

    error: str = Field(
        default="",
        description="Error message if any",
    )


class Write(LocalAction[WriteRequest, WriteResponse]):
    """
    Write the given content to a file.

    Note:
        This action will replace the existing content in the the file,
    and completely rewrite the file, if you want to edit a specific portion
    of the file use `edit` tool instead.
    """

    @include_cwd  # type: ignore
    def execute(self, request: WriteRequest, metadata: Dict) -> WriteResponse:
        try:
            filemanager = self.filemanagers.get(request.file_manager_id)
            (
                filemanager.recent
                if request.file_path is None
                else filemanager.open(path=request.file_path)
            ).write(text=request.text)
            return WriteResponse()
        except FileNotFoundError as e:
            return WriteResponse(error=f"File not found: {str(e)}")
        except PermissionError as e:
            return WriteResponse(error=f"Permission denied: {str(e)}")
        except OSError as e:
            return WriteResponse(error=f"OS error occurred: {str(e)}")



"""
Integrations manager for Composio SDK.

Usage:
    composio integrations [command] [options]
"""

import click

from composio.cli.context import Context, login_required, pass_context
from composio.cli.utils.decorators import handle_exceptions
from composio.cli.utils.helpfulcmd import HelpfulCmdBase
from composio.core.cls.did_you_mean import DYMGroup


class IntegrationsExamples(HelpfulCmdBase, DYMGroup):
    examples = [
        click.style("composio integrations", fg="green")
        + click.style(
            "                                # List all integrations\n", fg="black"
        ),
        click.style("composio integrations add --name GitHub", fg="green")
        + click.style(
            "              # Add a new integration named GitHub\n", fg="black"
        ),
        click.style("composio integrations remove 123", fg="green")
        + click.style("                # Remove integration with ID 123\n", fg="black"),
    ]


@click.group(name="integrations", invoke_without_command=True, cls=IntegrationsExamples)
@click.help_option("--help", "-h", "-help")
@handle_exceptions()
@login_required
@pass_context
def _integrations(context: Context) -> None:
    """List composio integrations for your account"""
    if context.click_ctx.invoked_subcommand:
        return

    integrations = context.client.integrations.get()
    context.console.print("[green]Showing integrations[/green]")
    for integration in integrations:
        context.console.print(f"• App: {integration.appName}")
        context.console.print(f"  ID: {integration.id}")


@_integrations.command(name="remove")
@click.help_option("--help")
@click.argument("id", type=str)
@handle_exceptions()
@login_required
@pass_context
def _remove(context: Context, id: str) -> None:
    """Remove integration with given ID"""
    context.client.integrations.remove(id=id)
    context.console.print(f"[green]Removed integration with {id=}[/green]")



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class InviteGuestToWorkspaceRequest(BaseModel):
    """Request schema for `InviteGuestToWorkspace`"""

    team_id: int = Field(
        ...,
        alias="team_id",
        description="Team ID (Workspace)",
    )
    email: str = Field(
        default=...,
        alias="email",
        description="Email",
    )
    can_edit_tags: bool = Field(
        default=...,
        alias="can_edit_tags",
        description="Can Edit Tags",
    )
    can_see_time_spent: bool = Field(
        default=...,
        alias="can_see_time_spent",
        description="Can See Time Spent",
    )
    can_see_time_estimated: bool = Field(
        default=...,
        alias="can_see_time_estimated",
        description="Can See Time Estimated",
    )
    can_create_views: bool = Field(
        default=...,
        alias="can_create_views",
        description="Can Create Views",
    )
    custom_role_id: int = Field(
        default=...,
        alias="custom_role_id",
        description="Custom Role Id",
    )


class InviteGuestToWorkspaceResponse(BaseModel):
    """Response schema for `InviteGuestToWorkspace`"""

    data: t.Dict[str, t.Any]


class InviteGuestToWorkspace(OpenAPIAction):
    """
    To invite a guest to your Workspace, use the "Invite User to Workspace"
    endpoint. Then, use specific endpoints to grant access to folders, lists,
    or tasks. Note: Available only for Enterprise Plan.
    """

    _tags = ["Guests"]
    _display_name = "invite_guest_to_workspace"
    _request_schema = InviteGuestToWorkspaceRequest
    _response_schema = InviteGuestToWorkspaceResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/guest"
    method = "post"
    operation_id = "Guests_inviteToWorkspace"
    action_identifier = "/team/{team_id}/guest_post"

    path_params = {"team_id": "team_id"}
    query_params = {}
    header_params = {}
    request_params = {
        "email": {"__alias": "email"},
        "can_edit_tags": {"__alias": "can_edit_tags"},
        "can_see_time_spent": {"__alias": "can_see_time_spent"},
        "can_see_time_estimated": {"__alias": "can_see_time_estimated"},
        "can_create_views": {"__alias": "can_create_views"},
        "custom_role_id": {"__alias": "custom_role_id"},
    }

    aliases = {}



import typing as t

from .action import Action


class Tool:
    """Abstraction for local tools."""

    @property
    def name(self) -> str:
        """Tool name."""
        return self.__class__.__name__.lower()

    def get_action(self, name: str) -> Action:
        """Get action object."""
        for action in self.actions():
            instance = action()
            if instance.get_tool_merged_action_name() != name:
                continue
            return instance
        raise ValueError(f"No action found with name `{name}`")

    def actions(self) -> t.List[t.Type[Action]]:
        raise NotImplementedError("This method should be overridden by subclasses.")

    def triggers(self) -> t.List:
        raise NotImplementedError("This method should be overridden by subclasses.")
